@@ -112,7 +112,7 @@ unsigned char get_algorithm(EVP_PKEY *key) {
   }
 }
 
-char *string_parser(char *str_orig, char delimiter, char *str_found) {
+static char *string_parser(char *str_orig, char delimiter, char *str_found) {
   char escape_char = '\\';
   int f = 0;
   char *p = str_orig;
@@ -149,7 +149,8 @@ X509_NAME *parse_name(const char *orig_name) {
     fprintf(stderr, "Name is too long!\n");
     return NULL;
   }
-  strcpy(name, orig_name);
+  strncpy(name, orig_name, sizeof(name));
+  name[sizeof(name) - 1] = 0;
 
   if(*name != '/' || name[strlen(name)-1] != '/') {
     fprintf(stderr, "Name does not start or does not end with '/'!\n");
@@ -255,7 +256,7 @@ void dump_data(const unsigned char *buf, unsigned int len, FILE *output, bool sp
         if(BIO_write(b64, buf, (int)len) <= 0) {
           fprintf(stderr, "Failed to write data in base64 format\n");
         }
-        BIO_flush(b64);
+        (void)BIO_flush(b64);
         BIO_free_all(b64);
       }
       return;
@@ -365,16 +366,7 @@ int get_slot_hex(enum enum_slot slot_enum) {
 }
 
 bool set_component(unsigned char *in_ptr, const BIGNUM *bn, int element_len) {
-  int real_len = BN_num_bytes(bn);
-
-  if(real_len > element_len) {
-    return false;
-  }
-  memset(in_ptr, 0, (size_t)(element_len - real_len));
-  in_ptr += element_len - real_len;
-  BN_bn2bin(bn, in_ptr);
-
-  return true;
+  return BN_bn2binpad(bn, in_ptr, element_len) == element_len;
 }
 
 bool prepare_rsa_signature(const unsigned char *in, unsigned int in_len, unsigned char *out, unsigned int *out_len, int nid) {
@@ -552,7 +544,7 @@ unsigned char get_pin_policy(enum enum_pin_policy policy) {
       return YKPIV_PINPOLICY_ALWAYS;
     case pin_policy__NULL:
     default:
-      return 0;
+      return YKPIV_PINPOLICY_DEFAULT;
   }
 }
 
@@ -566,7 +558,7 @@ unsigned char get_touch_policy(enum enum_touch_policy policy) {
       return YKPIV_TOUCHPOLICY_CACHED;
     case touch_policy__NULL:
     default:
-      return 0;
+      return YKPIV_TOUCHPOLICY_DEFAULT;
   }
 }
 
@@ -644,7 +636,7 @@ int SSH_write_X509(FILE *fp, X509 *x) {
       BIO_free_all(b64);
       break;
     }
-    BIO_flush(b64);
+    (void)BIO_flush(b64);
     BIO_free_all(b64);
 
     ret = 1;
