@@ -102,7 +102,7 @@ ykpiv_rc ykpiv_util_get_cardid(ykpiv_state *state, ykpiv_cardid *cardid) {
   size_t offs, cb_temp = 0;
   uint8_t tag = 0;
 
-  if (!cardid) return YKPIV_GENERIC_ERROR;
+  if (!cardid) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -152,7 +152,7 @@ ykpiv_rc ykpiv_util_set_cardid(ykpiv_state *state, const ykpiv_cardid *cardid) {
   uint8_t buf[sizeof(CHUID_TMPL)] = {0};
   size_t len = 0;
 
-  if (!state) return YKPIV_GENERIC_ERROR;
+  if (!state) return YKPIV_ARGUMENT_ERROR;
 
   if (!cardid) {
     if (PRNG_OK != _ykpiv_prng_generate(id, sizeof(id))) {
@@ -183,7 +183,7 @@ ykpiv_rc ykpiv_util_get_cccid(ykpiv_state *state, ykpiv_cccid *ccc) {
   uint8_t buf[CB_OBJ_MAX] = {0};
   unsigned long len = sizeof(buf);
 
-  if (!ccc) return YKPIV_GENERIC_ERROR;
+  if (!ccc) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -210,7 +210,7 @@ ykpiv_rc ykpiv_util_set_cccid(ykpiv_state *state, const ykpiv_cccid *ccc) {
   uint8_t buf[sizeof(CCC_TMPL)] = {0};
   size_t len = 0;
 
-  if (!state) return YKPIV_GENERIC_ERROR;
+  if (!state) return YKPIV_ARGUMENT_ERROR;
 
   if (!ccc) {
     if (PRNG_OK != _ykpiv_prng_generate(id, sizeof(id))) {
@@ -282,7 +282,7 @@ ykpiv_rc ykpiv_util_list_keys(ykpiv_state *state, uint8_t *key_count, ykpiv_key
     YKPIV_KEY_CARDAUTH
   };
 
-  if ((NULL == data) || (NULL == data_len) || (NULL == key_count)) { return YKPIV_GENERIC_ERROR; }
+  if ((NULL == data) || (NULL == data_len) || (NULL == key_count)) { return YKPIV_ARGUMENT_ERROR; }
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -355,7 +355,7 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_free(ykpiv_state *state, void *data) {
   if (!data) return YKPIV_OK;
-  if (!state || (!(state->allocator.pfn_free))) return YKPIV_GENERIC_ERROR;
+  if (!state || (!(state->allocator.pfn_free))) return YKPIV_ARGUMENT_ERROR;
 
   _ykpiv_free(state, data);
 
@@ -367,7 +367,7 @@ ykpiv_rc ykpiv_util_read_cert(ykpiv_state *state, uint8_t slot, uint8_t **data,
   uint8_t buf[CB_BUF_MAX] = {0};
   size_t cbBuf = sizeof(buf);
 
-  if ((NULL == data )|| (NULL == data_len)) return YKPIV_GENERIC_ERROR;
+  if ((NULL == data )|| (NULL == data_len)) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -428,7 +428,7 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
   size_t  cb_item = 0;
   uint8_t flags = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -456,7 +456,7 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
         memcpy(&flags, p_item, cb_item);
       }
       else {
-        if (state->verbose) { fprintf(stderr, "admin flags exist, but are incorrect size = %lu", (unsigned long)cb_item); }
+        DBG("admin flags exist, but are incorrect size = %lu", (unsigned long)cb_item);
       }
     }
   }
@@ -464,11 +464,11 @@ ykpiv_rc ykpiv_util_block_puk(ykpiv_state *state) {
   flags |= ADMIN_FLAGS_1_PUK_BLOCKED;
 
   if (YKPIV_OK != _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_ADMIN_FLAGS_1, (uint8_t*)&flags, sizeof(flags))) {
-    if (state->verbose) { fprintf(stderr, "could not set admin flags"); }
+    DBG("could not set admin flags");
   }
   else {
     if (YKPIV_OK != _write_metadata(state, TAG_ADMIN, data, cb_data)) {
-      if (state->verbose) { fprintf(stderr, "could not write admin metadata"); }
+      DBG("could not write admin metadata");
     }
   }
 
@@ -485,7 +485,7 @@ ykpiv_rc ykpiv_util_read_mscmap(ykpiv_state *state, ykpiv_container **containers
   size_t offs, len = 0;
   uint8_t *ptr = NULL;
 
-  if ((NULL == containers) || (NULL == n_containers)) { res = YKPIV_GENERIC_ERROR; goto Cleanup; }
+  if ((NULL == containers) || (NULL == n_containers)) { res = YKPIV_ARGUMENT_ERROR; goto Cleanup; }
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
@@ -544,7 +544,7 @@ ykpiv_rc ykpiv_util_write_mscmap(ykpiv_state *state, ykpiv_container *containers
     // if either containers or n_containers are non-zero, return an error,
     // that we only delete strictly when both are set properly
     if ((NULL != containers) || (0 != n_containers)) {
-      res = YKPIV_GENERIC_ERROR;
+      res = YKPIV_ARGUMENT_ERROR;
     }
     else {
       res = _ykpiv_save_object(state, YKPIV_OBJ_MSCMAP, NULL, 0);
@@ -591,7 +591,7 @@ ykpiv_rc ykpiv_util_read_msroots(ykpiv_state *state, uint8_t **data, size_t *dat
   size_t cbRealloc = 0;
   size_t offset = 0;
 
-  if (!data || !data_len) return YKPIV_GENERIC_ERROR;
+  if (!data || !data_len) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -690,7 +690,7 @@ ykpiv_rc ykpiv_util_write_msroots(ykpiv_state *state, uint8_t *data, size_t data
     // if either data or data_len are non-zero, return an error,
     // that we only delete strictly when both are set properly
     if ((NULL != data) || (0 != data_len)) {
-      res = YKPIV_GENERIC_ERROR;
+      res = YKPIV_ARGUMENT_ERROR;
     }
     else {
       // it should be sufficient to just delete the first object, though
@@ -743,7 +743,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   unsigned char data[1024] = {0};
   unsigned char templ[] = { 0, YKPIV_INS_GENERATE_ASYMMETRIC, 0, 0 };
   unsigned long recv_len = sizeof(data);
-  int sw;
+  int sw = 0;
   uint8_t *ptr_modulus = NULL;
   size_t  cb_modulus = 0;
   uint8_t *ptr_exp = NULL;
@@ -787,8 +787,8 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
           break;
       }
 
-      fprintf(stderr, sz_roca_format, state->serial, psz_msg);
-      yc_log_event(1, setting_roca.value ? YC_LOG_LEVEL_WARN : YC_LOG_LEVEL_ERROR, sz_roca_format, state->serial, psz_msg);
+      DBG(sz_roca_format, state->serial, psz_msg);
+      yc_log_event("YubiKey PIV Library", 1, setting_roca.value ? YC_LOG_LEVEL_WARN : YC_LOG_LEVEL_ERROR, sz_roca_format, state->serial, psz_msg);
 
       if (!setting_roca.value) {
         return YKPIV_NOT_SUPPORTED;
@@ -800,8 +800,8 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   case YKPIV_ALGO_RSA1024:
   case YKPIV_ALGO_RSA2048:
     if (!modulus || !modulus_len || !exp || !exp_len) {
-      if (state->verbose) { fprintf(stderr, "Invalid output parameter for RSA algorithm"); }
-      return YKPIV_GENERIC_ERROR;
+      DBG("Invalid output parameter for RSA algorithm");
+      return YKPIV_ARGUMENT_ERROR;
     }
     *modulus = NULL;
     *modulus_len = 0;
@@ -812,15 +812,15 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
   case  YKPIV_ALGO_ECCP256:
   case  YKPIV_ALGO_ECCP384:
     if (!point || !point_len) {
-      if (state->verbose) { fprintf(stderr, "Invalid output parameter for ECC algorithm"); }
-      return YKPIV_GENERIC_ERROR;
+      DBG("Invalid output parameter for ECC algorithm");
+      return YKPIV_ARGUMENT_ERROR;
     }
     *point = NULL;
     *point_len = 0;
     break;
 
   default:
-    if (state->verbose) { fprintf(stderr, "Invalid algorithm specified"); }
+    DBG("Invalid algorithm specified");
     return YKPIV_GENERIC_ERROR;
   }
 
@@ -837,7 +837,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
 
   if (in_data[4] == 0) {
     res = YKPIV_ALGORITHM_ERROR;
-    if (state->verbose) { fprintf(stderr, "Unexpected algorithm.\n"); }
+    DBG("Unexpected algorithm");
     goto Cleanup;
   }
 
@@ -855,41 +855,12 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     *in_ptr++ = touch_policy;
   }
 
-  if (YKPIV_OK != (res = _ykpiv_transfer_data(state, templ, in_data, (long)(in_ptr - in_data), data, &recv_len, &sw))) {
-    if (state->verbose) { fprintf(stderr, "Failed to communicate.\n"); }
+  if (YKPIV_OK != (res = _ykpiv_transfer_data(state, templ, in_data, (unsigned long)(in_ptr - in_data), data, &recv_len, &sw))) {
     goto Cleanup;
   }
-  else if (sw != SW_SUCCESS) {
-    if (state->verbose) { fprintf(stderr, "Failed to generate new key ("); }
-
-    if (sw == SW_ERR_INCORRECT_SLOT) {
-      res = YKPIV_KEY_ERROR;
-      if (state->verbose) { fprintf(stderr, "incorrect slot)\n"); }
-    }
-    else if (sw == SW_ERR_INCORRECT_PARAM) {
-      res = YKPIV_ALGORITHM_ERROR;
-
-      if (state->verbose) {
-        if (pin_policy != YKPIV_PINPOLICY_DEFAULT) {
-          fprintf(stderr, "pin policy not supported?)\n");
-        }
-        else if (touch_policy != YKPIV_TOUCHPOLICY_DEFAULT) {
-          fprintf(stderr, "touch policy not supported?)\n");
-        }
-        else {
-          fprintf(stderr, "algorithm not supported?)\n");
-        }
-      }
-    }
-    else if (sw == SW_ERR_SECURITY_STATUS) {
-      res = YKPIV_AUTHENTICATION_ERROR;
-      if (state->verbose) { fprintf(stderr, "not authenticated)\n"); }
-    }
-    else {
-      res = YKPIV_GENERIC_ERROR;
-      if (state->verbose) { fprintf(stderr, "error %x)\n", sw); }
-    }
-
+  res = ykpiv_translate_sw(sw);
+  if (res != YKPIV_OK) {
+    DBG("Failed to generate new key");
     goto Cleanup;
   }
 
@@ -898,7 +869,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     unsigned char *data_ptr = data + 2 + _ykpiv_get_length(data + 2, data + recv_len, &len);
 
     if (*data_ptr != TAG_RSA_MODULUS) {
-      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (modulus).\n"); }
+      DBG("Failed to parse public key structure (modulus).");
       res = YKPIV_PARSE_ERROR;
       goto Cleanup;
     }
@@ -906,7 +877,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     data_ptr++;
     offs = _ykpiv_get_length(data_ptr, data + recv_len, &len);
     if(!offs) {
-      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (modulus length).\n"); }
+      DBG("Failed to parse public key structure (modulus length).");
       res = YKPIV_PARSE_ERROR;
       goto Cleanup;
     }
@@ -914,7 +885,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
 
     cb_modulus = len;
     if (NULL == (ptr_modulus = _ykpiv_alloc(state, cb_modulus))) {
-      if (state->verbose) { fprintf(stderr, "Failed to allocate memory for modulus.\n"); }
+      DBG("Failed to allocate memory for modulus.");
       res = YKPIV_MEMORY_ERROR;
       goto Cleanup;
     }
@@ -924,7 +895,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     data_ptr += len;
 
     if (*data_ptr != TAG_RSA_EXP) {
-      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (public exponent).\n"); }
+      DBG("Failed to parse public key structure (public exponent).");
       res = YKPIV_PARSE_ERROR;
       goto Cleanup;
     }
@@ -932,7 +903,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     data_ptr++;
     offs = _ykpiv_get_length(data_ptr, data + recv_len, &len);
     if(!offs) {
-      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure (public exponent length).\n"); }
+      DBG("Failed to parse public key structure (public exponent length).");
       res = YKPIV_PARSE_ERROR;
       goto Cleanup;
     }
@@ -940,7 +911,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
 
     cb_exp = len;
     if (NULL == (ptr_exp = _ykpiv_alloc(state, cb_exp))) {
-      if (state->verbose) { fprintf(stderr, "Failed to allocate memory for public exponent.\n"); }
+      DBG("Failed to allocate memory for public exponent.");
       res = YKPIV_MEMORY_ERROR;
       goto Cleanup;
     }
@@ -968,20 +939,20 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     }
 
     if (*data_ptr++ != TAG_ECC_POINT) {
-      if (state->verbose) { fprintf(stderr, "Failed to parse public key structure.\n"); }
+      DBG("Failed to parse public key structure.");
       res = YKPIV_PARSE_ERROR;
       goto Cleanup;
     }
 
     if (*data_ptr++ != len) { /* the curve point should always be determined by the curve */
-      if (state->verbose) { fprintf(stderr, "Unexpected length.\n"); }
+      DBG("Unexpected length.");
       res = YKPIV_ALGORITHM_ERROR;
       goto Cleanup;
     }
 
     cb_point = len;
     if (NULL == (ptr_point = _ykpiv_alloc(state, cb_point))) {
-      if (state->verbose) { fprintf(stderr, "Failed to allocate memory for public point.\n"); }
+      DBG("Failed to allocate memory for public point.");
       res = YKPIV_MEMORY_ERROR;
       goto Cleanup;
     }
@@ -995,7 +966,7 @@ ykpiv_rc ykpiv_util_generate_key(ykpiv_state *state, uint8_t slot, uint8_t algor
     *point_len = cb_point;
   }
   else {
-    if (state->verbose) { fprintf(stderr, "Wrong algorithm.\n"); }
+    DBG("Wrong algorithm.");
     res = YKPIV_ALGORITHM_ERROR;
     goto Cleanup;
   }
@@ -1017,8 +988,8 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
   uint8_t *p_item = NULL;
   size_t cb_item = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
-  if (NULL == config) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
+  if (NULL == config) return YKPIV_ARGUMENT_ERROR;
 
   // initialize default values
 
@@ -1039,10 +1010,8 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
 
     if (YKPIV_OK == _get_metadata_item(data, cb_data, TAG_ADMIN_SALT, &p_item, &cb_item)) {
       if (config->mgm_type != YKPIV_CONFIG_MGM_MANUAL) {
-        if (state->verbose) {
-          fprintf(stderr, "conflicting types of mgm key administration configured\n");
-          config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
-        }
+        DBG("conflicting types of mgm key administration configured");
+        config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
       }
       else {
         config->mgm_type = YKPIV_CONFIG_MGM_DERIVED;
@@ -1051,9 +1020,7 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
 
     if (YKPIV_OK == _get_metadata_item(data, cb_data, TAG_ADMIN_TIMESTAMP, &p_item, &cb_item)) {
       if (CB_ADMIN_TIMESTAMP != cb_item)  {
-        if (state->verbose) {
-          fprintf(stderr, "pin timestamp in admin metadata is an invalid size");
-        }
+        DBG("pin timestamp in admin metadata is an invalid size");
       }
       else {
         memcpy(&(config->pin_last_changed), p_item, cb_item);
@@ -1071,21 +1038,22 @@ ykpiv_rc ykpiv_util_get_config(ykpiv_state *state, ykpiv_config *config) {
     }
 
     if (YKPIV_OK == _get_metadata_item(data, cb_data, TAG_PROTECTED_MGM, &p_item, &cb_item)) {
-      if(sizeof(config->mgm_key) == cb_item) {
+      if(sizeof(config->mgm_key) >= cb_item) {
         memcpy(config->mgm_key, p_item, cb_item);
+        config->mgm_len = cb_item;
         if (config->mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
-          if (state->verbose) fprintf(stderr, "conflicting types of mgm key administration configured - protected mgm exists\n");
-          config->mgm_type = YKPIV_CONFIG_MGM_PROTECTED;
+          DBG("conflicting types of mgm key administration configured - protected mgm exists");
+          config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
         }
       } else {
-        if (state->verbose) fprintf(stderr, "protected data contains mgm, but is the wrong size = %lu\n", (unsigned long)cb_item);
+        DBG("protected data contains mgm, but is the wrong size = %lu", (unsigned long)cb_item);
         config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
       }
     }
   }
   else {
     if (config->mgm_type == YKPIV_CONFIG_MGM_PROTECTED) {
-      if (state->verbose) fprintf(stderr, "admin data indicates protected mgm present, but the object cannot be read\n");
+      DBG("admin data indicates protected mgm present, but the object cannot be read");
       config->mgm_type = YKPIV_CONFIG_MGM_INVALID;
     }
   }
@@ -1103,7 +1071,7 @@ ykpiv_rc ykpiv_util_set_pin_last_changed(ykpiv_state *state) {
   size_t   cb_data = sizeof(data);
   time_t   tnow = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -1116,12 +1084,12 @@ ykpiv_rc ykpiv_util_set_pin_last_changed(ykpiv_state *state) {
   tnow = time(NULL);
 
   if (YKPIV_OK != (res = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_ADMIN_TIMESTAMP, (uint8_t*)&tnow, CB_ADMIN_TIMESTAMP))) {
-    if (state->verbose) fprintf(stderr, "could not set pin timestamp, err = %d\n", res);
+    DBG("could not set pin timestamp, err = %d", res);
   }
   else {
     if (YKPIV_OK != (res = _write_metadata(state, TAG_ADMIN, data, cb_data))) {
       /* Note: this can fail if authenticate() wasn't called previously - expected behavior */
-      if (state->verbose) fprintf(stderr, "could not write admin data, err = %d\n", res);
+      DBG("could not write admin data, err = %d", res);
     }
   }
 
@@ -1139,8 +1107,8 @@ ykpiv_rc ykpiv_util_get_derived_mgm(ykpiv_state *state, const uint8_t *pin, cons
   uint8_t  *p_item = NULL;
   size_t   cb_item = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
-  if ((NULL == pin) || (0 == pin_len) || (NULL == mgm)) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
+  if ((NULL == pin) || (0 == pin_len) || (NULL == mgm)) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
@@ -1149,13 +1117,13 @@ ykpiv_rc ykpiv_util_get_derived_mgm(ykpiv_state *state, const uint8_t *pin, cons
   if (YKPIV_OK == (res = _read_metadata(state, TAG_ADMIN, data, &cb_data))) {
     if (YKPIV_OK == (res = _get_metadata_item(data, cb_data, TAG_ADMIN_SALT, &p_item, &cb_item))) {
       if (cb_item != CB_ADMIN_SALT) {
-        if (state->verbose) fprintf(stderr, "derived mgm salt exists, but is incorrect size = %lu\n", (unsigned long)cb_item);
+        DBG("derived mgm salt exists, but is incorrect size = %lu", (unsigned long)cb_item);
         res = YKPIV_GENERIC_ERROR;
         goto Cleanup;
       }
-
-      if (PKCS5_OK != (p5rc = pkcs5_pbkdf2_sha1(pin, pin_len, p_item, cb_item, ITER_MGM_PBKDF2, mgm->data, member_size(ykpiv_mgm, data)))) {
-        if (state->verbose) fprintf(stderr, "pbkdf2 failure, err = %d\n", p5rc);
+      mgm->len = DES_LEN_3DES;
+      if (PKCS5_OK != (p5rc = pkcs5_pbkdf2_sha1(pin, pin_len, p_item, cb_item, ITER_MGM_PBKDF2, mgm->data, mgm->len))) {
+        DBG("pbkdf2 failure, err = %d", p5rc);
         res = YKPIV_GENERIC_ERROR;
         goto Cleanup;
       }
@@ -1175,28 +1143,29 @@ ykpiv_rc ykpiv_util_get_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   uint8_t  *p_item = NULL;
   size_t   cb_item = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
-  if (NULL == mgm) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
+  if (NULL == mgm) return YKPIV_ARGUMENT_ERROR;
 
   if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) return res;
   if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
 
   if (YKPIV_OK != (res = _read_metadata(state, TAG_PROTECTED, data, &cb_data))) {
-    if (state->verbose) fprintf(stderr, "could not read protected data, err = %d\n", res);
+    DBG("could not read protected data, err = %d", res);
     goto Cleanup;
   }
 
   if (YKPIV_OK != (res = _get_metadata_item(data, cb_data, TAG_PROTECTED_MGM, &p_item, &cb_item))) {
-    if (state->verbose) fprintf(stderr, "could not read protected mgm from metadata, err = %d\n", res);
+    DBG("could not read protected mgm from metadata, err = %d", res);
     goto Cleanup;
   }
 
-  if (cb_item != member_size(ykpiv_mgm, data)) {
-    if (state->verbose) fprintf(stderr, "protected data contains mgm, but is the wrong size = %lu\n", (unsigned long)cb_item);
+  if (cb_item > sizeof(mgm->data)) {
+    DBG("protected data contains mgm, but is the wrong size = %lu", (unsigned long)cb_item);
     res = YKPIV_AUTHENTICATION_ERROR;
     goto Cleanup;
   }
 
+  mgm->len = cb_item;
   memcpy(mgm->data, p_item, cb_item);
 
 Cleanup:
@@ -1208,13 +1177,42 @@ Cleanup:
 
 }
 
+ykpiv_rc ykpiv_util_update_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
+  ykpiv_rc res = YKPIV_OK;
+  uint8_t data[CB_BUF_MAX] = {0};
+  size_t cb_data = sizeof(data);
+
+  if (YKPIV_OK != (res = _ykpiv_begin_transaction(state))) goto Cleanup;
+  if (YKPIV_OK != (res = _ykpiv_ensure_application_selected(state))) goto Cleanup;
+
+  if (YKPIV_OK != (res = _read_metadata(state, TAG_PROTECTED, data, &cb_data))) {
+    cb_data = 0; /* set current metadata blob size to zero, we'll add to the blank blob */
+  }
+
+  if (YKPIV_OK != (res = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_PROTECTED_MGM, mgm->data, mgm->len))) {
+    DBG("could not set protected mgm item, err = %d", res);
+  }
+  else {
+    if (YKPIV_OK != (res = _write_metadata(state, TAG_PROTECTED, data, cb_data))) {
+      DBG("could not write protected data, err = %d", res);
+      goto Cleanup;
+    }
+  }
+
+Cleanup:
+
+  _ykpiv_end_transaction(state);
+  return res;
+}
+
 /* to set a generated mgm, pass NULL for mgm, or set mgm.data to all zeroes */
 ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   ykpiv_rc res = YKPIV_OK;
   ykpiv_rc ykrc = YKPIV_OK;
   prng_rc  prngrc = PRNG_OK;
   bool     fGenerate = false;
-  uint8_t  mgm_key[member_size(ykpiv_mgm, data)] = { 0 };
+  size_t   mgm_len = DES_LEN_3DES;
+  uint8_t  mgm_key[sizeof(mgm->data)] = { 0 };
   size_t   i = 0;
   uint8_t  data[CB_BUF_MAX] = { 0 };
   size_t   cb_data = sizeof(data);
@@ -1222,16 +1220,14 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   size_t   cb_item = 0;
   uint8_t  flags_1 = 0;
 
-  if (NULL == state) return YKPIV_GENERIC_ERROR;
+  if (NULL == state) return YKPIV_ARGUMENT_ERROR;
 
-  if (!mgm) {
-    fGenerate = true;
-  }
-  else {
-    fGenerate = true;
-    memcpy(mgm_key, mgm->data, sizeof(mgm_key));
+  fGenerate = true;
+  if (mgm) {
+    mgm_len = mgm->len;
+    memcpy(mgm_key, mgm->data, mgm->len);
 
-    for (i = 0; i < sizeof(mgm_key); i++) {
+    for (i = 0; i < mgm_len; i++) {
       if (mgm_key[i] != 0) {
         fGenerate = false;
         break;
@@ -1246,21 +1242,21 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
   do {
     if (fGenerate) {
       /* generate a new mgm key */
-      if (PRNG_OK != (prngrc = _ykpiv_prng_generate(mgm_key, sizeof(mgm_key)))) {
-        if (state->verbose) fprintf(stderr, "could not generate new mgm, err = %d\n", prngrc);
+      if (PRNG_OK != (prngrc = _ykpiv_prng_generate(mgm_key, mgm_len))) {
+        DBG("could not generate new mgm, err = %d", prngrc);
         res = YKPIV_RANDOMNESS_ERROR;
         goto Cleanup;
       }
     }
 
-    if (YKPIV_OK != (ykrc = ykpiv_set_mgmkey(state, mgm_key))) {
+    if (YKPIV_OK != (ykrc = ykpiv_set_mgmkey3(state, mgm_key, mgm_len, YKPIV_ALGO_AUTO, YKPIV_TOUCHPOLICY_AUTO))) {
       /*
       ** if _set_mgmkey fails with YKPIV_KEY_ERROR, it means the generated key is weak
       ** otherwise, log a warning, since the device mgm key is corrupt or we're in
       ** a state where we can't set the mgm key
       */
       if (YKPIV_KEY_ERROR != ykrc) {
-        if (state->verbose) fprintf(stderr, "could not set new derived mgm key, err = %d\n", ykrc);
+        DBG("could not set new derived mgm key, err = %d", ykrc);
         res = ykrc;
         goto Cleanup;
       }
@@ -1273,7 +1269,7 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
 
   /* set output mgm */
   if (mgm) {
-    memcpy(mgm->data, mgm_key, sizeof(mgm_key));
+    memcpy(mgm->data, mgm_key, mgm_len);
   }
 
   /* after this point, we've set the mgm key, so the function should succeed, regardless of being able to set the metadata */
@@ -1283,12 +1279,12 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
     cb_data = 0; /* set current metadata blob size to zero, we'll add to the blank blob */
   }
 
-  if (YKPIV_OK != (ykrc = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_PROTECTED_MGM, mgm_key, sizeof(mgm_key)))) {
-    if (state->verbose) fprintf(stderr, "could not set protected mgm item, err = %d\n", ykrc);
+  if (YKPIV_OK != (ykrc = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_PROTECTED_MGM, mgm_key, mgm_len))) {
+    DBG("could not set protected mgm item, err = %d", ykrc);
   }
   else {
     if (YKPIV_OK != (ykrc = _write_metadata(state, TAG_PROTECTED, data, cb_data))) {
-      if (state->verbose) fprintf(stderr, "could not write protected data, err = %d\n", ykrc);
+      DBG("could not write protected data, err = %d", ykrc);
       goto Cleanup;
     }
   }
@@ -1303,30 +1299,30 @@ ykpiv_rc ykpiv_util_set_protected_mgm(ykpiv_state *state, ykpiv_mgm *mgm) {
 
     if (YKPIV_OK != (ykrc = _get_metadata_item(data, cb_data, TAG_ADMIN_FLAGS_1, &p_item, &cb_item))) {
       /* flags are not set */
-      if (state->verbose) fprintf(stderr, "admin data exists, but flags are not present\n");
+      DBG("admin data exists, but flags are not present");
     }
 
     if (cb_item == sizeof(flags_1)) {
       memcpy(&flags_1, p_item, cb_item);
     }
     else {
-      if (state->verbose) fprintf(stderr, "admin data flags are an incorrect size = %lu\n", (unsigned long)cb_item);
+      DBG("admin data flags are an incorrect size = %lu", (unsigned long)cb_item);
     }
 
     /* remove any existing salt */
     if (YKPIV_OK != (ykrc = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_ADMIN_SALT, NULL, 0))) {
-      if (state->verbose) fprintf(stderr, "could not unset derived mgm salt, err = %d\n", ykrc);
+      DBG("could not unset derived mgm salt, err = %d", ykrc);
     }
   }
 
   flags_1 |= ADMIN_FLAGS_1_PROTECTED_MGM;
 
   if (YKPIV_OK != (ykrc = _set_metadata_item(data, &cb_data, CB_OBJ_MAX, TAG_ADMIN_FLAGS_1, &flags_1, sizeof(flags_1)))) {
-    if (state->verbose) fprintf(stderr, "could not set admin flags item, err = %d\n", ykrc);
+    DBG("could not set admin flags item, err = %d", ykrc);
   }
   else {
     if (YKPIV_OK != (ykrc = _write_metadata(state, TAG_ADMIN, data, cb_data))) {
-      if (state->verbose) fprintf(stderr, "could not write admin data, err = %d\n", ykrc);
+      DBG("could not write admin data, err = %d", ykrc);
       goto Cleanup;
     }
   }
@@ -1343,17 +1339,17 @@ Cleanup:
 
 ykpiv_rc ykpiv_util_reset(ykpiv_state *state) {
   unsigned char templ[] = {0, YKPIV_INS_RESET, 0, 0};
-  unsigned char data[0xff] = {0};
+  unsigned char data[256] = {0};
   unsigned long recv_len = sizeof(data);
   ykpiv_rc res;
   int sw;
 
   /* note: the reset function is only available when both pins are blocked. */
   res = ykpiv_transfer_data(state, templ, NULL, 0, data, &recv_len, &sw);
-  if (YKPIV_OK == res && SW_SUCCESS == sw) {
-     return YKPIV_OK;
+  if(res != YKPIV_OK) {
+    return res;
   }
-  return YKPIV_GENERIC_ERROR;
+  return ykpiv_translate_sw(sw);
 }
 
 uint32_t ykpiv_util_slot_object(uint8_t slot) {
@@ -1443,7 +1439,7 @@ static ykpiv_rc _write_certificate(ykpiv_state *state, uint8_t slot, uint8_t *da
     // if either data or data_len are non-zero, return an error,
     // that we only delete strictly when both are set properly
     if ((NULL != data) || (0 != data_len)) {
-      return YKPIV_GENERIC_ERROR;
+      return YKPIV_ARGUMENT_ERROR;
     }
 
     return _ykpiv_save_object(state, object_id, NULL, 0);
@@ -1498,7 +1494,7 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
   uint8_t tag_temp = 0;
   bool found = false;
 
-  if (!data || !pp_item || !pcb_item) return YKPIV_GENERIC_ERROR;
+  if (!data || !pp_item || !pcb_item) return YKPIV_ARGUMENT_ERROR;
 
   *pp_item = NULL;
   *pcb_item = 0;
@@ -1531,41 +1527,37 @@ static ykpiv_rc _get_metadata_item(uint8_t *data, size_t cb_data, uint8_t tag, u
 }
 
 ykpiv_rc ykpiv_util_parse_metadata(uint8_t *data, size_t data_len, ykpiv_metadata *metadata) {
-  uint8_t *p;
-  size_t cb;
+  uint8_t *p = 0;
+  size_t cb = 0;
+  uint32_t cnt = 0;
 
   ykpiv_rc rc = _get_metadata_item(data, data_len, YKPIV_METADATA_ALGORITHM_TAG, &p, &cb);
-  if(rc != YKPIV_OK)
-    return rc;
-  if(cb != 1)
-    return YKPIV_PARSE_ERROR;
-  metadata->algorithm = p[0];
+  if(rc == YKPIV_OK && cb == 1) {
+    metadata->algorithm = p[0];
+    cnt++;
+  }
 
   rc = _get_metadata_item(data, data_len, YKPIV_METADATA_POLICY_TAG, &p, &cb);
-  if(rc != YKPIV_OK)
-    return rc;
-  if(cb != 2)
-    return YKPIV_PARSE_ERROR;
-  metadata->pin_policy = p[0];
-  metadata->touch_policy = p[1];
+  if(rc == YKPIV_OK && cb == 2) {
+    metadata->pin_policy = p[0];
+    metadata->touch_policy = p[1];
+    cnt++;
+  }
 
   rc = _get_metadata_item(data, data_len, YKPIV_METADATA_ORIGIN_TAG, &p, &cb);
-  if(rc != YKPIV_OK)
-    return rc;
-  if(cb != 1)
-    return YKPIV_PARSE_ERROR;
-  metadata->origin = p[0];
+  if(rc == YKPIV_OK && cb == 1) {
+    metadata->origin = p[0];
+    cnt++;
+  }
 
   rc = _get_metadata_item(data, data_len, YKPIV_METADATA_PUBKEY_TAG, &p, &cb);
-  if(rc != YKPIV_OK)
-    return rc;
-  if(cb > sizeof(metadata->pubkey))
-    return YKPIV_PARSE_ERROR;
+  if(rc == YKPIV_OK && cb > 0 && cb <= sizeof(metadata->pubkey)) {
+    metadata->pubkey_len = cb;
+    memcpy(metadata->pubkey, p, cb);
+    cnt++;
+  }
 
-  metadata->pubkey_len = cb;
-  memcpy(metadata->pubkey, p, cb);
-
-  return YKPIV_OK;
+  return cnt ? YKPIV_OK : YKPIV_PARSE_ERROR;
 }
 
 /*
@@ -1586,7 +1578,7 @@ static ykpiv_rc _set_metadata_item(uint8_t *data, size_t *pcb_data, size_t cb_da
   uint8_t *p_next = NULL;
   long    cb_moved = 0; /* must be signed to have negative offsets */
 
-  if (!data || !pcb_data) return YKPIV_GENERIC_ERROR;
+  if (!data || !pcb_data) return YKPIV_ARGUMENT_ERROR;
 
   while (p_temp < (data + *pcb_data)) {
     tag_temp = *p_temp++;
@@ -1674,7 +1666,7 @@ static ykpiv_rc _read_metadata(ykpiv_state *state, uint8_t tag, uint8_t* data, s
   size_t offs;
   int obj_id = 0;
 
-  if (!data || !pcb_data || (CB_BUF_MAX > *pcb_data)) return YKPIV_GENERIC_ERROR;
+  if (!data || !pcb_data || (CB_BUF_MAX > *pcb_data)) return YKPIV_ARGUMENT_ERROR;
 
   switch (tag) {
   case TAG_ADMIN: obj_id = YKPIV_OBJ_ADMIN_DATA; break;
