@@ -31,6 +31,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdbool.h>
+#include <locale.h>
 #include <limits.h>
 #include <string.h>
 #include <sys/types.h>
@@ -53,6 +54,7 @@
 #include <openssl/rand.h>
 #include <openssl/rsa.h>
 #include <openssl/x509v3.h>
+#include <openssl/err.h>
 
 #include "cmdline.h"
 #include "../common/util.h"
@@ -64,7 +66,7 @@
 
 #define MAX_OID_LEN 19
 
-#define KEY_LEN 24
+#define KEY_LEN 32
 
 #define YKPIV_ATTESTATION_OID "1.3.6.1.4.1.41482.3"
 
@@ -125,9 +127,11 @@ static bool sign_data(ykpiv_state *state, const unsigned char *in, size_t len, u
     in = signinput;
     len = padlen;
   }
-  if(ykpiv_sign_data(state, in, len, out, out_len, algorithm, key) == YKPIV_OK) {
+  ykpiv_rc rc;
+  if((rc = ykpiv_sign_data(state, in, len, out, out_len, algorithm, key)) == YKPIV_OK) {
     return true;
   }
+  fprintf(stderr, "Failed signing data: %s.\n", ykpiv_strerror(rc));
   return false;
 }
 
@@ -142,7 +146,12 @@ struct internal_key {
   size_t oid_len;
 };
 
-int yk_rsa_meth_sign(int dtype, const unsigned char *m, unsigned int m_length,
+static int yk_rsa_meth_finish(RSA *rsa) {
+  free(RSA_meth_get0_app_data(RSA_get_method(rsa)));
+  return 1;
+}
+
+static int yk_rsa_meth_sign(int dtype, const unsigned char *m, unsigned int m_length,
     unsigned char *sigret, unsigned int *siglen, const RSA *rsa) {
   size_t yk_siglen = RSA_size(rsa);
   const RSA_METHOD *meth = RSA_get_method(rsa);
@@ -163,7 +172,11 @@ int yk_rsa_meth_sign(int dtype, const unsigned char *m, unsigned int m_length,
   return 0;
 }
 
-int yk_ec_meth_sign(int type, const unsigned char *dgst, int dlen,
+static void yk_ec_meth_finish(EC_KEY *ec) {
+  free(EC_KEY_get_ex_data(ec, ec_key_ex_data_idx));
+}
+
+static int yk_ec_meth_sign(int type, const unsigned char *dgst, int dlen,
     unsigned char *sig, unsigned int *siglen, const BIGNUM *kinv,
     const BIGNUM *r, EC_KEY *ec) {
   size_t yk_siglen = ECDSA_size(ec);
@@ -176,39 +189,53 @@ int yk_ec_meth_sign(int type, const unsigned char *dgst, int dlen,
   return 0;
 }
 
-static void wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_key,
+static EVP_PKEY* wrap_public_key(ykpiv_state *state, int algorithm, EVP_PKEY *public_key, 
     int key, const unsigned char *oid, size_t oid_len) {
-  static struct internal_key int_key;
-  int_key.state = state;
-  int_key.algorithm = algorithm;
-  int_key.key = key;
-  int_key.oid = oid;
-  int_key.oid_len = oid_len;
-  if(YKPIV_IS_RSA(algorithm)) {
+  struct internal_key *int_key = malloc(sizeof(struct internal_key));
+  int_key->state = state;
+  int_key->algorithm = algorithm;
+  int_key->key = key;
+  int_key->oid = oid;
+  int_key->oid_len = oid_len;
+  EVP_PKEY *pkey = EVP_PKEY_new();
+  if (YKPIV_IS_RSA(algorithm)) {
+    const RSA *pk = EVP_PKEY_get0_RSA(public_key);
     RSA_METHOD *meth = RSA_meth_dup(RSA_get_default_method());
-    RSA *rsa = EVP_PKEY_get0_RSA(public_key);
-    if(RSA_meth_set0_app_data(meth, &int_key) != 1) {
+    if(RSA_meth_set0_app_data(meth, int_key) != 1) {
       fprintf(stderr, "Failed to set RSA data\n");
     }
+    if(RSA_meth_set_finish(meth, yk_rsa_meth_finish) != 1) {
+      fprintf(stderr, "Failed to set RSA finish method\n");
+    }
     if(RSA_meth_set_sign(meth, yk_rsa_meth_sign) != 1) {
-      fprintf(stderr, "Failed to set RSA signature verification method\n");
+      fprintf(stderr, "Failed to set RSA sign method\n");
     }
-    if(RSA_set_method(rsa, meth) != 1) {
-      fprintf(stderr, "Failed to wrap RSA key\n");
+    RSA *sk = RSA_new();
+    RSA_set0_key(sk, BN_dup(RSA_get0_n(pk)), BN_dup(RSA_get0_e(pk)), NULL);
+    if(RSA_set_method(sk, meth) != 1) {
+      fprintf(stderr, "Failed to set RSA key method\n");
     }
-  } else {
-    EC_KEY *ec = EVP_PKEY_get0_EC_KEY(public_key);
+    EVP_PKEY_assign_RSA(pkey, sk);
+  }
+  else {
+    const EC_KEY *ec = EVP_PKEY_get0_EC_KEY(public_key);
     EC_KEY_METHOD *meth = EC_KEY_METHOD_new(EC_KEY_get_method(ec));
+    EC_KEY_METHOD_set_init(meth, NULL, yk_ec_meth_finish, NULL, NULL, NULL, NULL);
+    EC_KEY_METHOD_set_sign(meth, yk_ec_meth_sign, NULL, NULL);
+    EC_KEY *sk = EC_KEY_new();
+    EC_KEY_set_group(sk, EC_KEY_get0_group(ec));
+    EC_KEY_set_public_key(sk, EC_KEY_get0_public_key(ec));
     if (ec_key_ex_data_idx == -1)
-      ec_key_ex_data_idx = EC_KEY_get_ex_new_index(0, NULL, NULL, NULL, 0);
-    if(EC_KEY_set_ex_data(ec, ec_key_ex_data_idx, &int_key) != 1) {
+      ec_key_ex_data_idx = EC_KEY_get_ex_new_index(0, NULL, NULL, NULL, NULL);
+    if(EC_KEY_set_ex_data(sk, ec_key_ex_data_idx, int_key) != 1) {
       fprintf(stderr, "Failed to set EC data\n");
     }
-    EC_KEY_METHOD_set_sign(meth, yk_ec_meth_sign, NULL, NULL);
-    if(EC_KEY_set_method(ec, meth) != 1) {
+    if(EC_KEY_set_method(sk, meth) != 1) {
       fprintf(stderr, "Failed to wrap public EC key\n");
     }
+    EVP_PKEY_assign_EC_KEY(pkey, sk);
   }
+  return pkey;
 }
 #endif
 
@@ -295,7 +322,7 @@ static bool generate_key(ykpiv_state *state, enum enum_slot slot,
       }
       eckey = EC_KEY_new();
       group = EC_GROUP_new_by_curve_name(nid);
-      EC_GROUP_set_asn1_flag(group, nid);
+      EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
       if(EC_KEY_set_group(eckey, group) != 1) {
         fprintf(stderr, "Failed to set EC group.\n");
         goto generate_out;
@@ -423,21 +450,13 @@ static bool import_key(ykpiv_state *state, enum enum_key_format key_format,
 
   {
     unsigned char algorithm = get_algorithm(private_key);
-    unsigned char pp = YKPIV_PINPOLICY_DEFAULT;
-    unsigned char tp = YKPIV_TOUCHPOLICY_DEFAULT;
+    unsigned char pp = get_pin_policy(pin_policy);
+    unsigned char tp = get_touch_policy(touch_policy);
 
     if(algorithm == 0) {
       goto import_out;
     }
 
-    if(pin_policy != pin_policy__NULL) {
-        pp = get_pin_policy(pin_policy);
-    }
-
-    if(touch_policy != touch_policy__NULL) {
-      tp = get_touch_policy(touch_policy);
-    }
-
     if(YKPIV_IS_RSA(algorithm)) {
       RSA *rsa_private_key = EVP_PKEY_get1_RSA(private_key);
       unsigned char e[4] = {0};
@@ -679,11 +698,14 @@ static bool set_cardid(ykpiv_state *state, int verbose, int type) {
   return res == YKPIV_OK;
 }
 
-static int add_ext(STACK_OF(X509_EXTENSION) *exts, const char *oid, const char *name, const char *descr, const unsigned char *data, int len) {
-  int nid = OBJ_create(oid, name, descr);
+static X509_EXTENSION *create_ext(const char *oid, const char *name, const char *descr, const unsigned char *data, int len) {
+  int nid = OBJ_txt2nid(oid);
   if(nid <= 0) {
-    fprintf(stderr, "Failed creating %s extension object.\n", name);
-    return 0;
+    nid = OBJ_create(oid, name, descr);
+    if (nid <= 0) {
+      fprintf(stderr, "Failed creating %s extension object.\n", name);
+      return 0;
+    }
   }
   ASN1_OCTET_STRING *octets = ASN1_OCTET_STRING_new();
   if(!octets) {
@@ -697,9 +719,17 @@ static int add_ext(STACK_OF(X509_EXTENSION) *exts, const char *oid, const char *
   X509_EXTENSION *ext = X509_EXTENSION_create_by_NID(NULL, nid, 0, octets);
   if(!ext) {
     fprintf(stderr, "Failed creating %s extension.\n", name);
+  }
+  return ext;
+}
+
+static int add_ext(STACK_OF(X509_EXTENSION) * exts, const char *oid, const char *name, const char *descr, const unsigned char *data, int len)
+{
+  X509_EXTENSION *ext = create_ext(oid, name, descr, data, len);
+  if (!ext) {
     return 0;
   }
-  if(!sk_X509_EXTENSION_push(exts, ext)) {
+  if (!sk_X509_EXTENSION_push(exts, ext)) {
     fprintf(stderr, "Failed pushing %s extension.\n", name);
     return 0;
   }
@@ -708,7 +738,7 @@ static int add_ext(STACK_OF(X509_EXTENSION) *exts, const char *oid, const char *
 
 static bool request_certificate(ykpiv_state *state, enum enum_key_format key_format,
     const char *input_file_name, enum enum_slot slot, char *subject, enum enum_hash hash,
-    const char *output_file_name, int attestation) {
+    const char *output_file_name, int attest) {
   X509_REQ *req = NULL;
   X509_NAME *name = NULL;
   FILE *input_file = NULL;
@@ -744,7 +774,7 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
     goto request_out;
   }
 
-  if(attestation) {
+  if(attest) {
     unsigned char buf[YKPIV_OBJ_MAX_SIZE] = {0};
     size_t buflen = sizeof(buf);
     ykpiv_rc rc;
@@ -876,12 +906,15 @@ static bool request_certificate(ykpiv_state *state, enum enum_key_format key_for
 #else
   /* With opaque structures we can not touch whatever we want, but we need
    * to embed the sign_data function in the RSA/EC key structures  */
-  wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
+  EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  if(X509_REQ_sign(req, public_key, md) == 0) {
+  if(X509_REQ_sign(req, sk, md) == 0) {
     fprintf(stderr, "Failed signing request.\n");
+    ERR_print_errors_fp(stderr);
+    EVP_PKEY_free(sk);
     goto request_out;
   }
+  EVP_PKEY_free(sk);
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -901,20 +934,16 @@ request_out:
   if(output_file && output_file != stdout) {
     fclose(output_file);
   }
-  if(public_key) {
-    EVP_PKEY_free(public_key);
-  }
-  if(req) {
+  EVP_PKEY_free(public_key);
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  if(req) {
     if(req->sig_alg->parameter) {
       req->sig_alg->parameter = NULL;
     }
+  }
 #endif
-    X509_REQ_free(req);
-  }
-  if(name) {
-    X509_NAME_free(name);
-  }
+  X509_REQ_free(req);
+  X509_NAME_free(name);
   return ret;
 }
 
@@ -930,58 +959,91 @@ static const struct {
 
 static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_format,
     const char *input_file_name, enum enum_slot slot, char *subject, enum enum_hash hash,
-    const int *serial, int validDays, const char *output_file_name) {
-  FILE *input_file = NULL;
-  FILE *output_file = NULL;
+    const int *serial, int validDays, const char *output_file_name, int attest) {
   bool ret = false;
-  EVP_PKEY *public_key = NULL;
+  X509_EXTENSION *attestation = NULL;
+  X509_EXTENSION *attest_cert = NULL;
   X509 *x509 = NULL;
+  EVP_PKEY *public_key = NULL;
   X509_NAME *name = NULL;
-  const EVP_MD *md;
-  unsigned char algorithm;
-  int key = 0;
-  size_t oid_len;
-  const unsigned char *oid;
-  int nid;
   ASN1_INTEGER *sno = ASN1_INTEGER_new();
   BIGNUM *ser = NULL;
-#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
-  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN] = {0};
-  unsigned int digest_len;
-  unsigned int md_len;
-  unsigned char *signinput;
-  size_t len = 0;
-  ASN1_TYPE null_parameter;
-#endif
 
-  key = get_slot_hex(slot);
+  int key = get_slot_hex(slot);
 
-  input_file = open_file(input_file_name, key_file_mode(key_format, false));
-  output_file = open_file(output_file_name, key_file_mode(key_format, true));
+  FILE *input_file = open_file(input_file_name, key_file_mode(key_format, false));
+  FILE *output_file = open_file(output_file_name, key_file_mode(key_format, true));
   if(!input_file || !output_file) {
     goto selfsign_out;
   }
 
-  if(isatty(fileno(input_file))) {
-    fprintf(stderr, "Please paste the public key...\n");
+  if (attest) {
+    unsigned char buf[YKPIV_OBJ_MAX_SIZE] = {0};
+    size_t buflen = sizeof(buf);
+    ykpiv_rc rc = 0;
+
+    if ((rc = ykpiv_attest(state, key, buf, &buflen)) == YKPIV_OK) {
+      // Extract the public key for the request from the attestation
+      const unsigned char *ptr = buf;
+      X509 *cert = d2i_X509(NULL, &ptr, buflen);
+      if (cert) {
+        public_key = X509_get_pubkey(cert);
+        if (!public_key) {
+          fprintf(stderr, "Failed extracting public key from attestation.\n");
+        }
+        X509_free(cert);
+      } else {
+        fprintf(stderr, "Failed extracting attestation.\n");
+      }
+
+      attestation = create_ext(YKPIV_ATTESTATION_OID ".11", "ykpiv attestation", "Yubico PIV X.509 Attestation", buf, buflen);
+      if (!attestation) {
+        fprintf(stderr, "Failed creating attestation extension.\n");
+      }
+
+      unsigned char *pb = 0;
+      size_t pblen = 0;
+      if ((rc = ykpiv_util_read_cert(state, YKPIV_KEY_ATTESTATION, &pb, &pblen)) == YKPIV_OK && pblen > 0) {
+        attest_cert = create_ext(YKPIV_ATTESTATION_OID ".2", "ykpiv attest cert", "Yubico PIV Attestation Certificate", pb, pblen);
+        if (!attest_cert) {
+          fprintf(stderr, "Failed creating attestation certificate extension.\n");
+        }
+        ykpiv_util_free(state, pb);
+      }
+      else {
+        fprintf(stderr, "Failed reading attestation certificate: %s.\n", ykpiv_strerror(rc));
+      }
+    }
+    else {
+      fprintf(stderr, "Failed creating attestation: %s.\n", ykpiv_strerror(rc));
+    }
   }
 
-  if(key_format == key_format_arg_PEM) {
-    public_key = PEM_read_PUBKEY(input_file, NULL, NULL, NULL);
-    if(!public_key) {
-      fprintf(stderr, "Failed loading public key for certificate.\n");
+  if(!public_key) {
+    if(isatty(fileno(input_file))) {
+      fprintf(stderr, "Please paste the public key...\n");
+    }
+
+    if(key_format == key_format_arg_PEM) {
+      public_key = PEM_read_PUBKEY(input_file, NULL, NULL, NULL);
+      if(!public_key) {
+        fprintf(stderr, "Failed loading public key for certificate.\n");
+        goto selfsign_out;
+      }
+    } else {
+      fprintf(stderr, "Only PEM supported for public key input.\n");
       goto selfsign_out;
     }
-  } else {
-    fprintf(stderr, "Only PEM supported for public key input.\n");
-    goto selfsign_out;
   }
-  algorithm = get_algorithm(public_key);
+
+  unsigned char algorithm = get_algorithm(public_key);
   if(algorithm == 0) {
     goto selfsign_out;
   }
 
-  md = get_hash(hash, &oid, &oid_len);
+  size_t oid_len = 0;
+  const unsigned char *oid = 0;
+  const EVP_MD *md = get_hash(hash, &oid, &oid_len);
   if(md == NULL) {
     goto selfsign_out;
   }
@@ -1043,7 +1105,7 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     fprintf(stderr, "Failed setting certificate issuer.\n");
     goto selfsign_out;
   }
-  nid = get_hashnid(hash, algorithm);
+  int nid = get_hashnid(hash, algorithm);
   if(nid == 0) {
     goto selfsign_out;
   }
@@ -1087,14 +1149,29 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
         goto selfsign_out;
       }
     }
+
+    if(attestation && !X509_add_ext(x509, attestation, -1)) {
+      fprintf(stderr, "Failed adding attestation extension.\n");
+      goto selfsign_out;
+    }
+
+    if(attest_cert && !X509_add_ext(x509, attest_cert, -1)) {
+      fprintf(stderr, "Failed adding attestation certificate extension.\n");
+      goto selfsign_out;
+    }
   }
 
 #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
+  unsigned char digest[EVP_MAX_MD_SIZE + MAX_OID_LEN] = {0};
+  unsigned char *signinput;
+  size_t len = 0;
+
+  ASN1_TYPE null_parameter;
   null_parameter.type = V_ASN1_NULL;
   null_parameter.value.ptr = NULL;
 
-  md_len = (unsigned int)EVP_MD_size(md);
-  digest_len = sizeof(digest) - md_len;
+  unsigned int md_len = (unsigned int)EVP_MD_size(md);
+  unsigned int digest_len = sizeof(digest) - md_len;
 
   if(YKPIV_IS_RSA(algorithm)) {
     signinput = digest;
@@ -1132,12 +1209,16 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
 #else
   /* With opaque structures we can not touch whatever we want, but we need
    * to embed the sign_data function in the RSA/EC key structures  */
-  wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
+  EVP_PKEY *sk = wrap_public_key(state, algorithm, public_key, key, oid, oid_len);
 
-  if(X509_sign(x509, public_key, md) == 0) {
+  if(X509_sign(x509, sk, md) == 0) {
     fprintf(stderr, "Failed signing certificate.\n");
+    ERR_print_errors_fp(stderr);
+    EVP_PKEY_free(sk);
     goto selfsign_out;
   }
+
+  EVP_PKEY_free(sk);
 #endif
 
   if(key_format == key_format_arg_PEM) {
@@ -1148,7 +1229,7 @@ static bool selfsign_certificate(ykpiv_state *state, enum enum_key_format key_fo
     }
 
   } else {
-    fprintf(stderr, "Only PEM support available for certificate requests.\n");
+    fprintf(stderr, "Only PEM support available for certificates.\n");
   }
 
 selfsign_out:
@@ -1158,27 +1239,19 @@ selfsign_out:
   if(output_file && output_file != stdout) {
     fclose(output_file);
   }
+ #if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
   if(x509) {
-#if (OPENSSL_VERSION_NUMBER < 0x10100000L) || defined(LIBRESSL_VERSION_NUMBER)
-    if(x509->sig_alg->parameter) {
+   if(x509->sig_alg->parameter) {
       x509->sig_alg->parameter = NULL;
       x509->cert_info->signature->parameter = NULL;
     }
+  }
 #endif
-    X509_free(x509);
-  }
-  if(public_key) {
-    EVP_PKEY_free(public_key);
-  }
-  if(name) {
-    X509_NAME_free(name);
-  }
-  if(ser) {
-    BN_free(ser);
-  }
-  if(sno) {
-    ASN1_INTEGER_free(sno);
-  }
+  X509_free(x509);
+  EVP_PKEY_free(public_key);
+  X509_NAME_free(name);
+  BN_free(ser);
+  ASN1_INTEGER_free(sno);
   return ret;
 }
 
@@ -1212,7 +1285,7 @@ static bool verify_pin(ykpiv_state *state, const char *pin) {
 static bool change_pin(ykpiv_state *state, enum enum_action action, const char *pin,
     const char *new_pin) {
   const char *name = action == action_arg_changeMINUS_pin ? "pin" : "puk";
-  int (*op)(ykpiv_state *state, const char * puk, size_t puk_len,
+  ykpiv_rc (*op)(ykpiv_state *state, const char * puk, size_t puk_len,
             const char * new_pin, size_t new_pin_len, int *tries) = ykpiv_change_pin;
   size_t pin_len;
   size_t new_len;
@@ -1381,24 +1454,24 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
 
     mdctx = EVP_MD_CTX_create();
     if(EVP_DigestInit_ex(mdctx, md, NULL) != 1) {
-      fprintf(stderr, "failed to initialize digest operation\n");
+      fprintf(stderr, "Failed to initialize digest operation\n");
       goto out;
     }
     while(!feof(input_file)) {
       char buf[1024] = {0};
       size_t len = fread(buf, 1, 1024, input_file);
       if(EVP_DigestUpdate(mdctx, buf, len) != 1) {
-        fprintf(stderr, "failed to update digest data\n");
+        fprintf(stderr, "Failed to update digest data\n");
         goto out;
       }
     }
     if(EVP_DigestFinal_ex(mdctx, hashed, &hash_len) != 1) {
-      fprintf(stderr, "failed to finalize digest operation\n");
+      fprintf(stderr, "Failed to finalize digest operation\n");
       goto out;
     }
 
     if(verbosity) {
-      fprintf(stderr, "file hashed as: ");
+      fprintf(stderr, "File hashed as: ");
       dump_data(hashed, hash_len, stderr, true, format_arg_hex);
     }
     EVP_MD_CTX_destroy(mdctx);
@@ -1412,12 +1485,12 @@ static bool sign_file(ykpiv_state *state, const char *input, const char *output,
     unsigned char buf[1024] = {0};
     size_t len = sizeof(buf);
     if(!sign_data(state, hashed, hash_len, buf, &len, algo, key)) {
-      fprintf(stderr, "failed signing file\n");
+      fprintf(stderr, "Failed signing file\n");
       goto out;
     }
 
     if(verbosity) {
-      fprintf(stderr, "file signed as: ");
+      fprintf(stderr, "File signed as: ");
       dump_data(buf, len, stderr, true, format_arg_hex);
     }
     fwrite(buf, 1, len, output_file);
@@ -1456,7 +1529,7 @@ static void print_cert_info(ykpiv_state *state, enum enum_slot slot, const EVP_M
 
   fprintf(output, "Slot %x:\t", slot_name);
 
-  if(*ptr++ == 0x70) {
+  if(*ptr++ == TAG_CERT) {
     unsigned int md_len = sizeof(data);
     const ASN1_TIME *not_before, *not_after;
 
@@ -2058,7 +2131,9 @@ read_out:
 
 int main(int argc, char *argv[]) {
   struct gengetopt_args_info args_info;
+  const uint8_t mgm_algo[] = {YKPIV_ALGO_3DES, YKPIV_ALGO_AES128, YKPIV_ALGO_AES192, YKPIV_ALGO_AES256};
   ykpiv_state *state;
+  ykpiv_rc rc;
   int verbosity;
   enum enum_action action;
   unsigned int i;
@@ -2067,11 +2142,15 @@ int main(int argc, char *argv[]) {
   char pwbuf[128] = {0};
   char *password;
 
+  if (setlocale(LC_ALL, "") == NULL) {
+    fprintf(stderr, "Warning, unable to reset locale\n");
+  }
+
   if(cmdline_parser(argc, argv, &args_info) != 0) {
     return EXIT_FAILURE;
   }
 
-  verbosity = args_info.verbose_arg + (int)args_info.verbose_given;
+  verbosity = args_info.verbose_arg ? args_info.verbose_arg : (int)args_info.verbose_given;
   password = args_info.password_arg;
 
   for(i = 0; i < args_info.action_given; i++) {
@@ -2135,14 +2214,26 @@ int main(int argc, char *argv[]) {
     }
   }
 
-  if(ykpiv_init(&state, verbosity) != YKPIV_OK) {
-    fprintf(stderr, "Failed initializing library.\n");
+  /* openssl setup.. */
+#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
+  OpenSSL_add_all_algorithms();
+#else
+  OPENSSL_init_crypto(OPENSSL_INIT_LOAD_CONFIG, 0);
+#endif
+
+  if((rc = ykpiv_init(&state, verbosity)) != YKPIV_OK) {
+    fprintf(stderr, "Failed initializing library: %s.\n", ykpiv_strerror(rc));
     cmdline_parser_free(&args_info);
     return EXIT_FAILURE;
   }
 
-  if(ykpiv_connect(state, args_info.reader_arg) != YKPIV_OK) {
-    fprintf(stderr, "Failed to connect to yubikey.\nTry removing and reconnecting the device.\n");
+  if((rc = ykpiv_connect(state, args_info.reader_arg)) != YKPIV_OK) {
+    fprintf(stderr, "Failed to connect to yubikey: %s.\n", ykpiv_strerror(rc));
+    if (rc == YKPIV_PCSC_SERVICE_ERROR) {
+      fprintf(stderr, "Try restarting the PCSC subsystem.\n");
+    } else if (rc == YKPIV_PCSC_ERROR) {
+      fprintf(stderr, "Try removing and reconnecting the device.\n");
+    }
     ykpiv_done(state);
     cmdline_parser_free(&args_info);
     return EXIT_FAILURE;
@@ -2150,12 +2241,16 @@ int main(int argc, char *argv[]) {
 
   for(i = 0; i < args_info.action_given; i++) {
     action = *(args_info.action_arg + i);
+    if(verbosity) {
+      fprintf(stderr, "Now processing for action '%s'.\n",
+          cmdline_parser_action_values[action]);
+    }
     switch(action) {
       case action_arg_importMINUS_key:
       case action_arg_importMINUS_certificate:
         if(args_info.key_format_arg == key_format_arg_PKCS12 && !password) {
           if(verbosity) {
-            fprintf(stderr, "Asking for password since '%s' needs it.\n", cmdline_parser_action_values[action]);
+            fprintf(stderr, "Asking for password since action '%s' needs it.\n", cmdline_parser_action_values[action]);
           }
           if(!read_pw("Password", pwbuf, sizeof(pwbuf), false, args_info.stdin_input_flag)) {
             fprintf(stderr, "Failed to get password.\n");
@@ -2174,31 +2269,38 @@ int main(int argc, char *argv[]) {
       case action_arg_deleteMINUS_certificate:
       case action_arg_writeMINUS_object:
         if(!authed) {
-          unsigned char key[KEY_LEN] = {0};
-          size_t key_len = sizeof(key);
-          char keybuf[KEY_LEN*2+2] = {0}; /* one extra byte for potential \n */
-          char *key_ptr = args_info.key_arg;
           if(verbosity) {
             fprintf(stderr, "Authenticating since action '%s' needs that.\n", cmdline_parser_action_values[action]);
           }
-          if(args_info.key_given && args_info.key_orig == NULL) {
-            if(!read_pw("management key", keybuf, sizeof(keybuf), false, args_info.stdin_input_flag)) {
-              fprintf(stderr, "Failed to read management key from stdin,\n");
-              ykpiv_done(state);
-              cmdline_parser_free(&args_info);
-              return EXIT_FAILURE;
-            }
-            key_ptr = keybuf;
-          }
-          if(ykpiv_hex_decode(key_ptr, strlen(key_ptr), key, &key_len) != YKPIV_OK) {
-            fprintf(stderr, "Failed decoding key!\n");
+          ykpiv_config cfg = {0};
+          if((rc = ykpiv_util_get_config(state, &cfg)) != YKPIV_OK) {
+            fprintf(stderr, "Failed to get config metadata: %s.\n", ykpiv_strerror(rc));
             ykpiv_done(state);
             cmdline_parser_free(&args_info);
             return EXIT_FAILURE;
           }
-
-          if(ykpiv_authenticate(state, key) != YKPIV_OK) {
-            fprintf(stderr, "Failed authentication with the application.\n");
+          if(cfg.mgm_type != YKPIV_CONFIG_MGM_PROTECTED) {
+            char keybuf[KEY_LEN * 2 + 2] = {0}; /* one extra byte for potential \n */
+            char *key_ptr = args_info.key_arg;
+            if(args_info.key_given && args_info.key_orig == NULL) {
+              if(!read_pw("management key", keybuf, sizeof(keybuf), false, args_info.stdin_input_flag)) {
+                fprintf(stderr, "Failed to read management key from stdin,\n");
+                ykpiv_done(state);
+                cmdline_parser_free(&args_info);
+                return EXIT_FAILURE;
+              }
+              key_ptr = keybuf;
+            }
+            cfg.mgm_len = sizeof(cfg.mgm_key);
+            if((rc = ykpiv_hex_decode(key_ptr, strlen(key_ptr), cfg.mgm_key, &cfg.mgm_len)) != YKPIV_OK) {
+              fprintf(stderr, "Failed decoding key: %s.\n", ykpiv_strerror(rc));
+              ykpiv_done(state);
+              cmdline_parser_free(&args_info);
+              return EXIT_FAILURE;
+            }
+          }
+          if((rc = ykpiv_authenticate2(state, cfg.mgm_key, cfg.mgm_len)) != YKPIV_OK) {
+            fprintf(stderr, "Failed authentication with the application: %s.\n", ykpiv_strerror(rc));
             ykpiv_done(state);
             cmdline_parser_free(&args_info);
             return EXIT_FAILURE;
@@ -2209,7 +2311,7 @@ int main(int argc, char *argv[]) {
           authed = true;
         } else {
           if(verbosity) {
-            fprintf(stderr, "Skipping authentication for '%s' since it's already done.\n", cmdline_parser_action_values[action]);
+            fprintf(stderr, "Skipping authentication for action '%s' since it's already done.\n", cmdline_parser_action_values[action]);
           }
         }
         break;
@@ -2234,23 +2336,6 @@ int main(int argc, char *argv[]) {
           fprintf(stderr, "Action '%s' does not need authentication.\n", cmdline_parser_action_values[action]);
         }
     }
-  }
-
-
-  /* openssl setup.. */
-#if (OPENSSL_VERSION_NUMBER < 0x10100000L)
-  OpenSSL_add_all_algorithms();
-#endif
-
-
-  for(i = 0; i < args_info.action_given; i++) {
-    char new_keybuf[KEY_LEN*2+2] = {0}; /* one extra byte for potential \n */
-    char *new_mgm_key = args_info.new_key_arg;
-    action = *(args_info.action_arg + i);
-    if(verbosity) {
-      fprintf(stderr, "Now processing for action '%s'.\n",
-          cmdline_parser_action_values[action]);
-    }
     switch(action) {
       case action_arg_version:
         print_version(state, args_info.output_arg);
@@ -2264,33 +2349,44 @@ int main(int argc, char *argv[]) {
         }
         break;
       case action_arg_setMINUS_mgmMINUS_key:
-        if(!new_mgm_key) {
-          if(!read_pw("new management key", new_keybuf, sizeof(new_keybuf), true, args_info.stdin_input_flag)) {
-            fprintf(stderr, "Failed to read management key from stdin,\n");
-            ret = EXIT_FAILURE;
-            break;
+        {
+          char new_keybuf[KEY_LEN * 2 + 2] = {0}; /* one extra byte for potential \n */
+          char *new_mgm_key = args_info.new_key_arg;
+          if(!new_mgm_key) {
+            if(!read_pw("new management key", new_keybuf, sizeof(new_keybuf), true, args_info.stdin_input_flag)) {
+              fprintf(stderr, "Failed to read management key from stdin,\n");
+              ret = EXIT_FAILURE;
+              break;
+            }
+            new_mgm_key = new_keybuf;
           }
-          new_mgm_key = new_keybuf;
-        }
-        if(strlen(new_mgm_key) == (KEY_LEN * 2)){
-          unsigned char new_key[KEY_LEN] = {0};
-          size_t new_key_len = sizeof(new_key);
-          if(ykpiv_hex_decode(new_mgm_key, strlen(new_mgm_key), new_key, &new_key_len) != YKPIV_OK) {
-            fprintf(stderr, "Failed decoding new key!\n");
+          ykpiv_mgm new_key = {KEY_LEN};
+          if((rc = ykpiv_hex_decode(new_mgm_key, strlen(new_mgm_key), new_key.data, &new_key.len)) != YKPIV_OK) {
+            fprintf(stderr, "Failed decoding new key: %s.\n", ykpiv_strerror(rc));
             ret = EXIT_FAILURE;
-          } else if(ykpiv_set_mgmkey2(state, new_key, args_info.touch_policy_arg == touch_policy_arg_always ? 1 : 0) != YKPIV_OK) {
-            fprintf(stderr, "Failed setting the new key!");
+          } else if((rc = ykpiv_set_mgmkey3(state, new_key.data, new_key.len, mgm_algo[args_info.new_key_algo_arg],
+                        get_touch_policy(args_info.touch_policy_arg))) != YKPIV_OK) {
+            fprintf(stderr, "Failed setting the new key: %s.\n", ykpiv_strerror(rc));
             if(args_info.touch_policy_arg != touch_policy__NULL) {
-              fprintf(stderr, " Maybe touch policy is not supported on this key?");
+              fprintf(stderr, " Maybe this touch policy or algorithm is not supported on this key?");
             }
             fprintf(stderr, "\n");
             ret = EXIT_FAILURE;
           } else {
             fprintf(stderr, "Successfully set new management key.\n");
+            ykpiv_config config = {0};
+            if((rc = ykpiv_util_get_config(state, &config)) != YKPIV_OK) {
+              fprintf(stderr, "Failed reading configuration metadata: %s.\n", ykpiv_strerror(rc));
+            } else {
+              if (config.mgm_type == YKPIV_CONFIG_MGM_PROTECTED) {
+                if((rc = ykpiv_util_update_protected_mgm(state, &new_key)) != YKPIV_OK) {
+                  fprintf(stderr, "Failed updating pin-protected management key metadata: %s.\n", ykpiv_strerror(rc));
+                } else {
+                  fprintf(stderr, "Successfully updated pin-protected management key metadata.\n");
+                }
+              }
+            }
           }
-        } else {
-          fprintf(stderr, "The new management key has to be exactly %d hexadecimal characters.\n", KEY_LEN * 2);
-          ret = EXIT_FAILURE;
         }
         break;
       case action_arg_reset:
@@ -2407,7 +2503,7 @@ int main(int argc, char *argv[]) {
         if(selfsign_certificate(state, args_info.key_format_arg, args_info.input_arg,
               args_info.slot_arg, args_info.subject_arg, args_info.hash_arg,
               args_info.serial_given ? &args_info.serial_arg : NULL, args_info.valid_days_arg,
-              args_info.output_arg) == false) {
+              args_info.output_arg, args_info.attestation_flag) == false) {
           ret = EXIT_FAILURE;
         } else {
           fprintf(stderr, "Successfully generated a new self signed certificate.\n");
@@ -2484,7 +2580,6 @@ int main(int argc, char *argv[]) {
         verbosity)) {
       fprintf(stderr, "Signature successful!\n");
     } else {
-      fprintf(stderr, "Failed signing!\n");
       ret = EXIT_FAILURE;
     }
   }
