@@ -477,6 +477,20 @@ static CK_RV _get_coa(ykcs11_x509_t **certs, piv_obj_id_t obj, CK_ATTRIBUTE_PTR
     data = b_tmp;
     break;
 
+  case CKA_COPYABLE:
+    DBG("COPYABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_DESTROYABLE:
+    DBG("DESTROYABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_TRUE;
+    data = b_tmp;
+    break;
+
   case CKA_TRUSTED:
     DBG("TRUSTED");
     len = sizeof(CK_BBOOL);
@@ -587,6 +601,20 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_COPYABLE:
+    DBG("COPYABLE"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_DESTROYABLE:
+    DBG("DESTROYABLE"); // Always true
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_TRUE;
+    data = b_tmp;
+    break;
+
   case CKA_NEVER_EXTRACTABLE:
     DBG("NEVER_EXTRACTABLE"); // Always true
     len = sizeof(CK_BBOOL);
@@ -601,6 +629,13 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_ENCRYPT:
+    DBG("ENCRYPT"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_DECRYPT:
     DBG("DECRYPT"); // Default empy
     len = sizeof(CK_BBOOL);
@@ -608,6 +643,20 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_WRAP:
+    DBG("WRAP"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_WRAP_WITH_TRUSTED:
+    DBG("WRAP_WITH_TRUSTED"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_UNWRAP:
     DBG("UNWRAP"); // Default empty
     len = sizeof(CK_BBOOL);
@@ -622,6 +671,20 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_VERIFY:
+    DBG("VERIFY"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_VERIFY_RECOVER:
+    DBG("VERIFY_RECOVER"); // Always false
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_DERIVE:
     DBG("DERIVE"); // Default false
     len = sizeof(CK_BBOOL);
@@ -631,16 +694,9 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
 
   case CKA_MODULUS:
     DBG("MODULUS");
-    len = sizeof(b_tmp);
+    len = do_get_key_size(s->pkeys[piv_objects[obj].sub_id]);
 
-    // Make sure that this is an RSA key
-    ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
-    if (ul_tmp == CKK_VENDOR_DEFINED)
-      return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_RSA)
-      return CKR_ATTRIBUTE_TYPE_INVALID;
-
-    if ((rv = do_get_modulus(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
+    if ((rv = do_get_modulus(s->pkeys[piv_objects[obj].sub_id], b_tmp, len)) != CKR_OK)
       return rv;
     data = b_tmp;
     break;
@@ -653,7 +709,7 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
     if (ul_tmp == CKK_VENDOR_DEFINED)
       return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_ECDSA)
+    if (ul_tmp != CKK_EC)
       return CKR_ATTRIBUTE_TYPE_INVALID;
 
     if ((rv = do_get_public_key(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
@@ -670,7 +726,7 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
     if (ul_tmp == CKK_VENDOR_DEFINED)
       return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_ECDSA)
+    if (ul_tmp != CKK_EC)
       return CKR_ATTRIBUTE_TYPE_INVALID;
 
     if ((rv = do_get_curve_parameters(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
@@ -682,22 +738,15 @@ static CK_RV get_proa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
   case CKA_MODULUS_BITS:
     DBG("MODULUS BITS");
     len = sizeof(CK_ULONG);
-    ul_tmp = do_get_key_size(s->pkeys[piv_objects[obj].sub_id]);
+    ul_tmp = do_get_key_bits(s->pkeys[piv_objects[obj].sub_id]);
     data = (CK_BYTE_PTR) &ul_tmp;
     break;
 
   case CKA_PUBLIC_EXPONENT:
     DBG("PUBLIC EXPONENT");
-    len = sizeof(CK_ULONG);
+    len = sizeof(F4) - 1;
 
-    // Make sure that this is an RSA key
-    ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
-    if (ul_tmp == CKK_VENDOR_DEFINED)
-      return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_RSA)
-      return CKR_ATTRIBUTE_TYPE_INVALID;
-
-    if ((rv = do_get_public_exponent(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
+    if ((rv = do_get_public_exponent(s->pkeys[piv_objects[obj].sub_id], b_tmp, len)) != CKR_OK)
       return rv;
     data = b_tmp;
     break;
@@ -775,6 +824,7 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
   CK_BYTE     b_tmp[1024] = {0};
   CK_ULONG    ul_tmp;
   CK_ULONG    len = 0;
+  CK_RV       rv;
   DBG("For public key object %u, get ", obj);
 
   switch (template->type) {
@@ -792,6 +842,13 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_ALWAYS_AUTHENTICATE:
+    DBG("ALWAYS_AUTHENTICATE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_PRIVATE:
     DBG("PRIVATE");
     len = sizeof(CK_BBOOL);
@@ -799,6 +856,48 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_SENSITIVE:
+    DBG("SENSITIVE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_ALWAYS_SENSITIVE:
+    DBG("ALWAYS_SENSITIVE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_EXTRACTABLE:
+    DBG("EXTRACTABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_TRUE;
+    data = b_tmp;
+    break;
+
+  case CKA_NEVER_EXTRACTABLE:
+    DBG("NEVER_EXTRACTABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_COPYABLE:
+    DBG("COPYABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_DESTROYABLE:
+    DBG("DESTROYABLE");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_TRUE;
+    data = b_tmp;
+    break;
+
   case CKA_LABEL:
     DBG("LABEL");
     len = strlen(piv_objects[obj].label);
@@ -842,13 +941,41 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
-  case CKA_VERIFY: // TODO: what about verify recover ?
+  case CKA_DECRYPT:
+    DBG("DECRYPT");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_SIGN:
+    DBG("SIGN");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_SIGN_RECOVER:
+    DBG("SIGN_RECOVER");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_VERIFY:
     DBG("VERIFY");
     len = sizeof(CK_BBOOL);
     b_tmp[0] = pubkey_objects[piv_objects[obj].sub_id].verify;
     data = b_tmp;
     break;
 
+  case CKA_VERIFY_RECOVER:
+    DBG("VERIFY_RECOVER");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_WRAP:
     DBG("WRAP");
     len = sizeof(CK_BBOOL);
@@ -856,6 +983,20 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     data = b_tmp;
     break;
 
+  case CKA_WRAP_WITH_TRUSTED:
+    DBG("WRAP_WITH_TRUSTED");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
+  case CKA_UNWRAP:
+    DBG("UNWRAP");
+    len = sizeof(CK_BBOOL);
+    b_tmp[0] = CK_FALSE;
+    data = b_tmp;
+    break;
+
   case CKA_DERIVE:
     DBG("DERIVE");
     len = sizeof(CK_BBOOL);
@@ -871,11 +1012,11 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
     if (ul_tmp == CKK_VENDOR_DEFINED)
       return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_ECDSA)
+    if (ul_tmp != CKK_EC)
       return CKR_ATTRIBUTE_TYPE_INVALID;
 
-    if (do_get_public_key(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len) != CKR_OK)
-      return CKR_FUNCTION_FAILED;
+    if ((rv = do_get_public_key(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
+      return rv;
     data = b_tmp;
     break;
 
@@ -888,51 +1029,37 @@ static CK_RV get_puoa(ykcs11_slot_t *s, piv_obj_id_t obj, CK_ATTRIBUTE_PTR templ
     ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
     if (ul_tmp == CKK_VENDOR_DEFINED)
       return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_ECDSA)
+    if (ul_tmp != CKK_EC)
       return CKR_ATTRIBUTE_TYPE_INVALID;
 
-    if (do_get_curve_parameters(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len) != CKR_OK)
-      return CKR_FUNCTION_FAILED;
+    if ((rv = do_get_curve_parameters(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len)) != CKR_OK)
+      return rv;
 
     data = b_tmp;
     break;
 
   case CKA_MODULUS:
     DBG("MODULUS");
-    len = sizeof(b_tmp);
+    len = do_get_key_size(s->pkeys[piv_objects[obj].sub_id]);
 
-    // Make sure that this is an RSA key
-    ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
-    if (ul_tmp == CKK_VENDOR_DEFINED)
-      return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_RSA)
-      return CKR_ATTRIBUTE_TYPE_INVALID;
-
-    if (do_get_modulus(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len) != CKR_OK)
-      return CKR_FUNCTION_FAILED;
+    if ((rv = do_get_modulus(s->pkeys[piv_objects[obj].sub_id], b_tmp, len)) != CKR_OK)
+      return rv;
     data = b_tmp;
     break;
 
   case CKA_MODULUS_BITS:
     DBG("MODULUS BITS");
     len = sizeof(CK_ULONG);
-    ul_tmp = do_get_key_size(s->pkeys[piv_objects[obj].sub_id]);
+    ul_tmp = do_get_key_bits(s->pkeys[piv_objects[obj].sub_id]);
     data = (CK_BYTE_PTR) &ul_tmp;
     break;
 
   case CKA_PUBLIC_EXPONENT:
     DBG("PUBLIC EXPONENT");
-    len = sizeof(CK_ULONG);
+    len = sizeof(F4) - 1;
 
-    // Make sure that this is an RSA key
-    ul_tmp = do_get_key_type(s->pkeys[piv_objects[obj].sub_id]); // Getting the info from the pubk
-    if (ul_tmp == CKK_VENDOR_DEFINED)
-      return CKR_FUNCTION_FAILED;
-    if (ul_tmp != CKK_RSA)
-      return CKR_ATTRIBUTE_TYPE_INVALID;
-
-    if (do_get_public_exponent(s->pkeys[piv_objects[obj].sub_id], b_tmp, &len) != CKR_OK)
-      return CKR_FUNCTION_FAILED;
+    if ((rv = do_get_public_exponent(s->pkeys[piv_objects[obj].sub_id], b_tmp, len)) != CKR_OK)
+      return rv;
     data = b_tmp;
     break;
 
@@ -1498,15 +1625,19 @@ CK_RV check_create_cert(CK_ATTRIBUTE_PTR templ, CK_ULONG n,
     switch (templ[i].type) {
     case CKA_CLASS:
       // Technically redundant check
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_CERTIFICATE)
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_CERTIFICATE) {
+        DBG("CKA_CLASS muste be CKO_CERTIFICATE");
         return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
 
       break;
 
     case CKA_ID:
       has_id = CK_TRUE;
-      if (find_cert_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ)
+      if (find_cert_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ) {
+        DBG("CKA_ID is not a valid PIV object id");
         return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
 
       *id = *((CK_BYTE_PTR)templ[i].pValue);
       break;
@@ -1519,11 +1650,14 @@ CK_RV check_create_cert(CK_ATTRIBUTE_PTR templ, CK_ULONG n,
       break;
 
     case CKA_TOKEN:
+    case CKA_COPYABLE:
+    case CKA_DESTROYABLE:
     case CKA_LABEL:
     case CKA_SUBJECT:
     case CKA_ISSUER:
     case CKA_CERTIFICATE_TYPE:
     case CKA_PRIVATE:
+    case CKA_SENSITIVE:
     case CKA_SERIAL_NUMBER:
     case CKA_NAME_HASH_ALGORITHM:
     case CKA_HASH_OF_SUBJECT_PUBLIC_KEY:
@@ -1531,7 +1665,7 @@ CK_RV check_create_cert(CK_ATTRIBUTE_PTR templ, CK_ULONG n,
       break;
 
     default:
-      DBG("Invalid %lx", templ[i].type);
+      DBG("Invalid attribute type 0x%lx", templ[i].type);
       return CKR_ATTRIBUTE_TYPE_INVALID;
     }
   }
@@ -1559,22 +1693,25 @@ CK_RV check_create_ec_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
   for (i = 0; i < n; i++) {
     switch (templ[i].type) {
     case CKA_CLASS:
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY)
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY) {
+        DBG("CKA_CLASS muste be CKO_PRIVATE_KEY");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       break;
 
     case CKA_KEY_TYPE:
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKK_ECDSA)
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKK_EC) {
+        DBG("CKA_KEY_TYPE muste be CKK_EC");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       break;
 
     case CKA_ID:
       has_id = CK_TRUE;
-      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ)
+      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ) {
+        DBG("CKA_ID is not a valid PIV object id");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       *id = *((CK_BYTE_PTR)templ[i].pValue);
       break;
 
@@ -1588,7 +1725,20 @@ CK_RV check_create_ec_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       has_params = CK_TRUE;
       ec_params = (CK_BYTE_PTR)templ[i].pValue;
       ec_params_len = templ[i].ulValueLen;
+      break;
 
+    case CKA_TOKEN:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_TOKEN must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_PRIVATE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_PRIVATE must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
       break;
 
     case CKA_SENSITIVE:
@@ -1605,6 +1755,76 @@ CK_RV check_create_ec_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       }
       break;
 
+    case CKA_COPYABLE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_COPYABLE must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_DESTROYABLE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_DESTROYABLE must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_ENCRYPT:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_ENCRYPT must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_SIGN:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_SIGN must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_SIGN_RECOVER:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_SIGN_RECOVER must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_VERIFY:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_VERIFY must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_VERIFY_RECOVER:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_VERIFY_RECOVER must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_DERIVE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_DERIVE must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_WRAP:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_WRAP must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_UNWRAP:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_UNWRAP must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
     case CKA_ALWAYS_AUTHENTICATE:
       if (*((CK_BBOOL *)templ[i].pValue) == CK_TRUE) {
         if (*pin_policy != YKPIV_PINPOLICY_DEFAULT &&
@@ -1658,15 +1878,14 @@ CK_RV check_create_ec_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       *pin_policy = b_tmp;
       break;
 
-    case CKA_TOKEN:
+    case CKA_DECRYPT:
     case CKA_LABEL:
     case CKA_SUBJECT:
-    case CKA_DERIVE:
       // Ignore other attributes
       break;
 
     default:
-      DBG("Invalid %lx", templ[i].type);
+      DBG("Invalid attribute type 0x%lx", templ[i].type);
       return CKR_ATTRIBUTE_TYPE_INVALID;
     }
   }
@@ -1712,64 +1931,78 @@ CK_RV check_create_rsa_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
   for (i = 0; i < n; i++) {
     switch (templ[i].type) {
     case CKA_CLASS:
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY)
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKO_PRIVATE_KEY) {
+        DBG("CKA_CLASS muste be CKO_PRIVATE_KEY");
         return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
 
+    case CKA_KEY_TYPE:
+      if (*((CK_ULONG_PTR)templ[i].pValue) != CKK_RSA) {
+        DBG("CKA_KEY_TYPE muste be CKK_RSA");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
       break;
 
     case CKA_ID:
       has_id = CK_TRUE;
-      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ)
+      if (find_pvtk_object(*((CK_BYTE_PTR)templ[i].pValue)) == PIV_INVALID_OBJ) {
+        DBG("CKA_ID is not a valid PIV object id");
         return CKR_ATTRIBUTE_VALUE_INVALID;
-
+      }
       *id = *((CK_BYTE_PTR)templ[i].pValue);
       break;
 
-    case CKA_KEY_TYPE:
-      if (*((CK_ULONG_PTR)templ[i].pValue) != CKK_RSA)
-        return CKR_ATTRIBUTE_VALUE_INVALID;
-
-      break;
-
     case CKA_PUBLIC_EXPONENT:
       has_e = CK_TRUE;
-      if (templ[i].ulValueLen != 3 || memcmp((CK_BYTE_PTR)templ[i].pValue, F4, 3) != 0)
+      if (templ[i].ulValueLen != 3 || memcmp((CK_BYTE_PTR)templ[i].pValue, F4, 3) != 0) {
+        DBG("CKA_PUBLIC_EXPONENT must be 0x010001");
         return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
       break;
 
     case CKA_PRIME_1:
       has_p = CK_TRUE;
       *p = (CK_BYTE_PTR)templ[i].pValue;
       *p_len = templ[i].ulValueLen;
-
       break;
 
     case CKA_PRIME_2:
       has_q = CK_TRUE;
       *q = (CK_BYTE_PTR)templ[i].pValue;
       *q_len = templ[i].ulValueLen;
-
       break;
 
     case CKA_EXPONENT_1:
       has_dp = CK_TRUE;
       *dp = (CK_BYTE_PTR)templ[i].pValue;
       *dp_len = templ[i].ulValueLen;
-
       break;
 
     case CKA_EXPONENT_2:
       has_dq = CK_TRUE;
       *dq = (CK_BYTE_PTR)templ[i].pValue;
       *dq_len = templ[i].ulValueLen;
-
       break;
 
     case CKA_COEFFICIENT:
       has_qinv = CK_TRUE;
       *qinv = (CK_BYTE_PTR)templ[i].pValue;
       *qinv_len = templ[i].ulValueLen;
+      break;
 
+    case CKA_TOKEN:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_TOKEN must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_PRIVATE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_PRIVATE must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
       break;
 
     case CKA_SENSITIVE:
@@ -1786,6 +2019,69 @@ CK_RV check_create_rsa_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       }
       break;
 
+    case CKA_ENCRYPT:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_ENCRYPT must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_DECRYPT:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_DECRYPT must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_SIGN:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_SIGN must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_SIGN_RECOVER:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_SIGN_RECOVER must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_VERIFY:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_VERIFY must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_VERIFY_RECOVER:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_VERIFY_RECOVER must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_DERIVE:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_TRUE) {
+        DBG("CKA_DERIVE must be TRUE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_WRAP:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_WRAP must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
+    case CKA_UNWRAP:
+      if (*((CK_BBOOL *)templ[i].pValue) != CK_FALSE) {
+        DBG("CKA_UNWRAP must be FALSE or omitted");
+        return CKR_ATTRIBUTE_VALUE_INVALID;
+      }
+      break;
+
     case CKA_ALWAYS_AUTHENTICATE:
       if (*((CK_BBOOL *)templ[i].pValue) == CK_TRUE) {
         if (*pin_policy != YKPIV_PINPOLICY_DEFAULT &&
@@ -1839,15 +2135,13 @@ CK_RV check_create_rsa_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       *pin_policy = b_tmp;
       break;
 
-    case CKA_TOKEN:
     case CKA_LABEL:
     case CKA_SUBJECT:
-    case CKA_DERIVE:
       // Ignore other attributes
       break;
 
     default:
-      DBG("Invalid %lx", templ[i].type);
+      DBG("Invalid attribute type 0x%lx", templ[i].type);
       return CKR_ATTRIBUTE_TYPE_INVALID;
     }
   }
@@ -1858,16 +2152,21 @@ CK_RV check_create_rsa_key(CK_ATTRIBUTE_PTR templ, CK_ULONG n, CK_BYTE_PTR id,
       has_q == CK_FALSE ||
       has_dp == CK_FALSE ||
       has_dq == CK_FALSE ||
-      has_qinv == CK_FALSE)
+      has_qinv == CK_FALSE) {
+    DBG("Missing RSA components");
     return CKR_TEMPLATE_INCOMPLETE;
+  }
 
-  if (*p_len != 64 && *p_len != 128)
+  if (*p_len != 64 && *p_len != 128) {
+    DBG("Invalid RSA component lengths");
     return CKR_ATTRIBUTE_VALUE_INVALID;
-
+  }
 
   if (*q_len != *p_len || *dp_len > *p_len ||
-      *dq_len > *p_len || *qinv_len > *p_len)
+      *dq_len > *p_len || *qinv_len > *p_len) {
+    DBG("Invalid RSA component lengths");
     return CKR_ATTRIBUTE_VALUE_INVALID;
+  }
 
   return CKR_OK;
 }
