@@ -49,7 +49,7 @@ CK_RV do_rsa_encrypt(ykcs11_pkey_t *key, int padding, const ykcs11_md_t* oaep_md
                      unsigned char *oaep_label, CK_ULONG oaep_label_len,
                      CK_BYTE_PTR data, CK_ULONG data_len, CK_BYTE_PTR enc, CK_ULONG_PTR enc_len) {
 
-  if (EVP_PKEY_base_id(key) != EVP_PKEY_RSA) {
+  if (!key || EVP_PKEY_base_id(key) != EVP_PKEY_RSA) { // EVP_PKEY_base_id doesn't handle NULL
     return CKR_KEY_TYPE_INCONSISTENT;
   }
 
@@ -110,7 +110,7 @@ CK_RV do_store_cert(CK_BYTE_PTR data, CK_ULONG len, ykcs11_x509_t **cert) {
   const unsigned char *p = data; // Mandatory temp variable required by OpenSSL
   unsigned long       offs, cert_len;
 
-  if (*p == 0x70) {
+  if (*p == TAG_CERT) {
     // The certificate is in "PIV" format 0x70 len 0x30 len ...
     p++;
     offs = get_length(p, data + len, &cert_len);
@@ -141,7 +141,7 @@ CK_RV do_generate_ec_key(int curve_name, ykcs11_pkey_t **pkey) {
   if(group == NULL) {
     return CKR_HOST_MEMORY;
   }
-  EC_GROUP_set_asn1_flag(group, curve_name);
+  EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
   eckey = EC_KEY_new();
   if(eckey == NULL) {
     rv = CKR_HOST_MEMORY;
@@ -177,7 +177,7 @@ CK_RV do_create_ec_key(CK_BYTE_PTR point, CK_ULONG point_len, int curve_name, yk
   EC_GROUP *group = EC_GROUP_new_by_curve_name(curve_name);
   if(group == NULL)
     return CKR_HOST_MEMORY;
-  EC_GROUP_set_asn1_flag(group, curve_name);
+  EC_GROUP_set_asn1_flag(group, OPENSSL_EC_NAMED_CURVE);
   eckey = EC_KEY_new();
   if(eckey == NULL) {
     rv = CKR_HOST_MEMORY;
@@ -200,6 +200,7 @@ CK_RV do_create_ec_key(CK_BYTE_PTR point, CK_ULONG point_len, int curve_name, yk
     rv = CKR_GENERAL_ERROR;
     goto create_ec_cleanup;
   }
+  EVP_PKEY_free(*pkey);
   *pkey = EVP_PKEY_new();
   if(*pkey == NULL) {
     rv = CKR_HOST_MEMORY;
@@ -241,6 +242,7 @@ CK_RV do_create_rsa_key(CK_BYTE_PTR mod, CK_ULONG mod_len, CK_BYTE_PTR exp, CK_U
     rv = CKR_GENERAL_ERROR;
     goto create_rsa_cleanup;
   }
+  EVP_PKEY_free(*pkey);
   *pkey = EVP_PKEY_new();
   if(*pkey == NULL) {
     rv = CKR_HOST_MEMORY;
@@ -357,7 +359,7 @@ CK_RV do_create_empty_cert(CK_BYTE_PTR in, CK_ULONG in_len, CK_ULONG algorithm,
     goto create_empty_cert_cleanup;
   }
 
-  if (len > *out_len) {
+  if ((CK_ULONG)len > *out_len) {
     rv = CKR_BUFFER_TOO_SMALL;
     goto create_empty_cert_cleanup;
   }
@@ -480,121 +482,140 @@ CK_RV do_store_pubk(ykcs11_x509_t *cert, ykcs11_pkey_t **key) {
 
 }
 
+static int OBJ_get_nid(const char *oid, const char *name, const char *descr) {
+  int nid = OBJ_txt2nid(oid);
+  if (nid <= 0) {
+    nid = OBJ_create(oid, name, descr);
+  }
+  return nid;
+}
+
+CK_RV do_parse_attestation(ykcs11_x509_t *cert, CK_BYTE_PTR pin_policy, CK_BYTE_PTR touch_policy) {
+
+  int nid = OBJ_get_nid(YKPIV_OID_USAGE_POLICY, "KeyUsagePolicy", "Yubico PIV key usage policy");
+  if(nid < 0)
+    return CKR_FUNCTION_FAILED;
+
+  int pos = X509_get_ext_by_NID(cert, nid, -1);
+  if (pos < 0)
+    return CKR_FUNCTION_FAILED;
+
+  X509_EXTENSION *ext = X509_get_ext(cert, pos);
+  if (ext == NULL)
+    return CKR_FUNCTION_FAILED;
+
+  ASN1_OCTET_STRING *oct = X509_EXTENSION_get_data(ext);
+  if (oct == NULL)
+    return CKR_FUNCTION_FAILED;
+
+  const unsigned char *p = ASN1_STRING_get0_data(oct);
+  *pin_policy = p[0];
+  *touch_policy = p[1];
+
+  return CKR_OK;
+}
+
 CK_KEY_TYPE do_get_key_type(ykcs11_pkey_t *key) {
 
-  switch (EVP_PKEY_base_id(key)) {
-  case EVP_PKEY_RSA:
-    return CKK_RSA;
+  if(key) { // EVP_PKEY_base_id doesn't handle NULL
+    switch (EVP_PKEY_base_id(key)) {
+    case EVP_PKEY_RSA:
+      return CKK_RSA;
 
-  case EVP_PKEY_EC:
-    return CKK_ECDSA;
-
-  default:
-    return CKK_VENDOR_DEFINED; // Actually an error
+    case EVP_PKEY_EC:
+      return CKK_EC;
+    }
   }
+  return CKK_VENDOR_DEFINED; // Actually an error
+}
+
+CK_ULONG do_get_key_bits(ykcs11_pkey_t *key) {
+  return EVP_PKEY_bits(key);
 }
 
 CK_ULONG do_get_key_size(ykcs11_pkey_t *key) {
-  return EVP_PKEY_bits(key);
+  return EVP_PKEY_size(key);
 }
 
 CK_ULONG do_get_signature_size(ykcs11_pkey_t *key) {
 
-  switch (EVP_PKEY_base_id(key)) {
-  case EVP_PKEY_RSA:
-    return EVP_PKEY_size(key);
-  case EVP_PKEY_EC:
-    switch(EVP_PKEY_bits(key)) {
-    case 256:
-      return 64;
-    case 384:
-      return 96;
+  if(key) { // EVP_PKEY_base_id doesn't handle NULL
+    switch (EVP_PKEY_base_id(key)) {
+    case EVP_PKEY_RSA:
+      return EVP_PKEY_size(key);
+    case EVP_PKEY_EC:
+      switch(EVP_PKEY_bits(key)) {
+      case 256:
+        return 64;
+      case 384:
+        return 96;
+      }
     }
   }
-  return 0; // Actually an error
+  return 0;
 }
 
 CK_BYTE do_get_key_algorithm(ykcs11_pkey_t *key) {
 
-  switch (EVP_PKEY_base_id(key)) {
-  case EVP_PKEY_RSA:
-    switch(EVP_PKEY_bits(key)) {
-    case 1024:
-      return YKPIV_ALGO_RSA1024;
-    case 2048:
-      return YKPIV_ALGO_RSA2048;
-    }
-  case EVP_PKEY_EC:
-    switch(EVP_PKEY_bits(key)) {
-    case 256:
-      return YKPIV_ALGO_ECCP256;
-    case 384:
-      return YKPIV_ALGO_ECCP384;
+  if(key) { // EVP_PKEY_base_id doesn't handle NULL
+    switch (EVP_PKEY_base_id(key)) {
+    case EVP_PKEY_RSA:
+      switch(EVP_PKEY_bits(key)) {
+      case 1024:
+        return YKPIV_ALGO_RSA1024;
+      case 2048:
+        return YKPIV_ALGO_RSA2048;
+      }
+    case EVP_PKEY_EC:
+      switch(EVP_PKEY_bits(key)) {
+      case 256:
+        return YKPIV_ALGO_ECCP256;
+      case 384:
+        return YKPIV_ALGO_ECCP384;
+      }
     }
   }
-  return 0; // Actually an error
+  return 0;
 }
 
-CK_RV do_get_modulus(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len) {
-  RSA *rsa = NULL;
+CK_RV do_get_modulus(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len) {
+  const RSA *rsa = NULL;
   const BIGNUM *n = NULL;
-  CK_RV rv;
 
-  rsa = EVP_PKEY_get0_RSA(key);
+  rsa = key ? EVP_PKEY_get0_RSA(key) : 0;
   if (rsa == NULL)
-    return CKR_FUNCTION_FAILED;
+    return CKR_ATTRIBUTE_TYPE_INVALID;
 
   RSA_get0_key(rsa, &n, NULL, NULL);
-  if ((CK_ULONG)BN_num_bytes(n) > *len) {
-    rv = CKR_BUFFER_TOO_SMALL;
-    goto get_mod_cleanup;
-  }
 
-  *len = (CK_ULONG)BN_bn2bin(n, data);
+  if(BN_bn2binpad(n, data, len) < 0)
+    return CKR_DATA_LEN_RANGE;
 
   return CKR_OK;
-get_mod_cleanup:
-  if(n != NULL) {
-    BN_free(n);
-  }
-  return rv;
 }
 
-CK_RV do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len) {
+CK_RV do_get_public_exponent(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG len) {
 
-  RSA *rsa = NULL;
+  const RSA *rsa = NULL;
   const BIGNUM *bn_e;
-  CK_RV rv;
 
-  rsa = EVP_PKEY_get0_RSA(key);
+  rsa = key ? EVP_PKEY_get0_RSA(key) : 0;
   if (rsa == NULL)
-    return CKR_FUNCTION_FAILED;
+    return CKR_ATTRIBUTE_TYPE_INVALID;
 
   RSA_get0_key(rsa, NULL, &bn_e, NULL);
-  if ((CK_ULONG)BN_num_bytes(bn_e) > *len) {
-    rv = CKR_BUFFER_TOO_SMALL;
-    goto get_pubexp_cleanup;
-  }
 
-  *len = (CK_ULONG)BN_bn2bin(bn_e, data);
+  if(BN_bn2binpad(bn_e, data, len) < 0)
+    return CKR_DATA_LEN_RANGE;
+
   return CKR_OK;
-get_pubexp_cleanup:
-  if(bn_e != NULL) {
-    BN_free(bn_e);
-  }
-  return rv;
 }
 
-/* #include <stdio.h> */
-/* #include <openssl/err.h> */
-/*   ERR_load_crypto_strings(); */
-/* //SSL_load_error_strings(); */
-/*   fprintf(stderr, "ERROR %s\n", ERR_error_string(ERR_get_error(), NULL)); */
 CK_RV do_get_public_key(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len) {
-  RSA *rsa = NULL;
+  const RSA *rsa = NULL;
   unsigned char *p;
 
-  EC_KEY *eck = NULL;
+  const EC_KEY *eck = NULL;
   const EC_GROUP *ecg; // Alternative solution is to get i2d_PUBKEY and manually offset
   const EC_POINT *ecp;
   point_conversion_form_t pcf = POINT_CONVERSION_UNCOMPRESSED;
@@ -614,14 +635,6 @@ CK_RV do_get_public_key(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len)
       return CKR_FUNCTION_FAILED;
     }
 
-    // TODO: this is the correct thing to do so that we strip out the exponent
-    // OTOH we also need a function to get the exponent out with CKA_PUBLIC_EXPONENT
-    /*BN_bn2bin(rsa->n, data);
-     *len = 256;*/
-
-    /* fprintf(stderr, "Public key is: \n"); */
-    /* dump_hex(data, *len, stderr, CK_TRUE); */
-
     break;
 
   case EVP_PKEY_EC:
@@ -649,7 +662,7 @@ CK_RV do_get_public_key(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len)
 }
 
 CK_RV do_get_curve_parameters(ykcs11_pkey_t *key, CK_BYTE_PTR data, CK_ULONG_PTR len) {
-  EC_KEY *eck = NULL;
+  const EC_KEY *eck = NULL;
   const EC_GROUP *ecg;
   unsigned char *p;
 
@@ -703,7 +716,7 @@ CK_RV do_apply_DER_encoding_to_ECSIG(CK_BYTE_PTR signature, CK_ULONG_PTR signatu
     goto adete_out;
   }
 
-  if (len > buf_size) {
+  if ((CK_ULONG)len > buf_size) {
     rv = CKR_BUFFER_TOO_SMALL;
     goto adete_out;
   }
@@ -728,25 +741,6 @@ adete_out:
   return rv;
 }
 
-static int BN_bn2bin_fixed(const BIGNUM *bn, CK_BYTE_PTR out, CK_ULONG len) {
-
-  CK_BYTE buf[1024] = {0};
-  int actual = BN_bn2bin(bn, buf);
-  if(actual <= 0)
-    return actual;
-  if(actual < len) {
-    memset(out,  0, len - actual);
-    memcpy(out + len - actual, buf, actual);
-  } else {
-    for(CK_ULONG i = 0; i < actual - len; i++) {
-      if(buf[i])
-        return -1; // Non-zero byte would have been lost
-    }
-    memcpy(out, buf + actual - len, len);
-  }
-  return len;
-}
-
 CK_RV do_strip_DER_encoding_from_ECSIG(CK_BYTE_PTR data, CK_ULONG len, CK_ULONG sig_len) {
   CK_RV rv;
   const CK_BYTE *p = data;
@@ -758,18 +752,17 @@ CK_RV do_strip_DER_encoding_from_ECSIG(CK_BYTE_PTR data, CK_ULONG len, CK_ULONG
   const BIGNUM *x, *y;
   ECDSA_SIG_get0(sig, &x, &y);
 
-  if(BN_bn2bin_fixed(x, data, sig_len / 2) <= 0) {
+  if(BN_bn2binpad(x, data, sig_len / 2) <= 0) {
     rv = CKR_DATA_INVALID;
     goto strip_der_cleanup;
   }
 
-  if(BN_bn2bin_fixed(y, data + sig_len / 2, sig_len / 2) <= 0) {
+  if(BN_bn2binpad(y, data + sig_len / 2, sig_len / 2) <= 0) {
     rv = CKR_DATA_INVALID;
     goto strip_der_cleanup;
   }
 
-  ECDSA_SIG_free(sig);
-  return CKR_OK;
+  rv = CKR_OK;
 strip_der_cleanup:
   ECDSA_SIG_free(sig);
   return rv;
