@@ -79,35 +79,35 @@ static void get_functions() {
 
 }
 
-static void init_connection() {
-  asrt(funcs->C_Initialize(NULL), CKR_OK, "INITIALIZE");
-  CK_SLOT_ID pSlotList;
-  CK_ULONG pulCount = 16;
-  asrt(funcs->C_GetSlotList(true, &pSlotList, &pulCount), CKR_OK, "GETSLOTLIST");
-}
-
 static void test_lib_info() {
   dprintf(0, "TEST START: test_lib_info()\n");
 
-  const CK_CHAR_PTR MANUFACTURER_ID = "Yubico (www.yubico.com)";
-  const CK_CHAR_PTR YKCS11_DESCRIPTION = "PKCS#11 PIV Library (SP-800-73)";
+  const CK_CHAR_PTR MANUFACTURER_ID = (const CK_CHAR_PTR)"Yubico (www.yubico.com)";
+  const CK_CHAR_PTR YKCS11_DESCRIPTION = (const CK_CHAR_PTR)"PKCS#11 PIV Library (SP-800-73)";
   const CK_ULONG CRYPTOKI_VERSION_MAJ = 2;
   const CK_ULONG CRYPTOKI_VERSION_MIN = 40;
 
   CK_INFO info;
   asrt(funcs->C_Initialize(NULL), CKR_OK, "INITIALIZE");
   asrt(funcs->C_GetInfo(&info), CKR_OK, "GET_INFO");
-  asrt(strncmp(info.manufacturerID, MANUFACTURER_ID, strlen(MANUFACTURER_ID)), 0, "MANUFACTURER");
+  asrt(strncmp((const char*)info.manufacturerID, (const char*)MANUFACTURER_ID, strlen((const char*)MANUFACTURER_ID)), 0, "MANUFACTURER");
   asrt(info.cryptokiVersion.major, CRYPTOKI_VERSION_MAJ, "CK_MAJ");
   asrt(info.cryptokiVersion.minor, CRYPTOKI_VERSION_MIN, "CK_MIN");
   asrt(info.libraryVersion.major, YKCS11_VERSION_MAJOR, "LIB_MAJ");
   asrt(info.libraryVersion.minor, ((YKCS11_VERSION_MINOR * 10) + YKCS11_VERSION_PATCH), "LIB_MIN");
-  asrt(strncmp(info.libraryDescription, YKCS11_DESCRIPTION, strlen(YKCS11_DESCRIPTION)), 0, "LIB_DESC");
+  asrt(strncmp((const char*)info.libraryDescription, (const char*)YKCS11_DESCRIPTION, strlen((const char*)YKCS11_DESCRIPTION)), 0, "LIB_DESC");
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
   dprintf(0, "TEST END: test_lib_info()\n");
 }
 
 #if HW_TESTS
+static void init_connection() {
+  asrt(funcs->C_Initialize(NULL), CKR_OK, "INITIALIZE");
+  CK_SLOT_ID pSlotList[16];
+  CK_ULONG pulCount = 16;
+  asrt(funcs->C_GetSlotList(true, pSlotList, &pulCount), CKR_OK, "GETSLOTLIST");
+}
+
 static void test_initalize() {
   dprintf(0, "TEST START: test_initalize()\n");
   asrt(funcs->C_Initialize(NULL), CKR_OK, "INITIALIZE");
@@ -140,7 +140,7 @@ static int test_token_info() {
   asrt(info.ulSessionCount, 0, "SESSION_COUNT");
   asrt(info.ulMaxRwSessionCount, 16, "MAX_RW_SESSION_COUNT");
   asrt(info.ulRwSessionCount, 0, "RW_SESSION_COUNT");
-  asrt(info.ulMaxPinLen, 48, "MAX_PIN_LEN");
+  asrt(info.ulMaxPinLen, 64, "MAX_PIN_LEN");
   asrt(info.ulMinPinLen, 6, "MIN_PIN_LEN");
   asrt(info.ulTotalPublicMemory, -1, "TOTAL_PUB_MEM");
   asrt(info.ulFreePublicMemory, -1, "FREE_PUB_MEM");
@@ -177,31 +177,31 @@ static void test_mechanism_list_and_info() {
   CK_ULONG              i;
 
   static const CK_MECHANISM_TYPE token_mechanisms[] = {
-    CKM_RSA_PKCS_KEY_PAIR_GEN,
-    CKM_RSA_PKCS,
-    CKM_RSA_PKCS_PSS,
-    CKM_RSA_PKCS_OAEP,
-    CKM_RSA_X_509,
-    CKM_SHA1_RSA_PKCS,
-    CKM_SHA256_RSA_PKCS,
-    CKM_SHA384_RSA_PKCS,
-    CKM_SHA512_RSA_PKCS,
-    CKM_SHA1_RSA_PKCS_PSS,
-    CKM_SHA256_RSA_PKCS_PSS,
-    CKM_SHA384_RSA_PKCS_PSS,
-    CKM_SHA512_RSA_PKCS_PSS,
-    CKM_EC_KEY_PAIR_GEN,
-    CKM_ECDSA,
-    CKM_ECDSA_SHA1,
-    CKM_ECDSA_SHA224,
-    CKM_ECDSA_SHA256,
-    CKM_ECDSA_SHA384,
-    CKM_ECDH1_DERIVE,
-    CKM_SHA_1,
-    CKM_SHA256,
-    CKM_SHA384,
-    CKM_SHA512
-  };
+      CKM_RSA_PKCS_KEY_PAIR_GEN,
+      CKM_RSA_PKCS,
+      CKM_RSA_PKCS_PSS,
+      CKM_RSA_PKCS_OAEP,
+      CKM_RSA_X_509,
+      CKM_SHA1_RSA_PKCS,
+      CKM_SHA256_RSA_PKCS,
+      CKM_SHA384_RSA_PKCS,
+      CKM_SHA512_RSA_PKCS,
+      CKM_SHA1_RSA_PKCS_PSS,
+      CKM_SHA256_RSA_PKCS_PSS,
+      CKM_SHA384_RSA_PKCS_PSS,
+      CKM_SHA512_RSA_PKCS_PSS,
+      CKM_EC_KEY_PAIR_GEN,
+      CKM_ECDSA,
+      CKM_ECDSA_SHA1,
+      CKM_ECDSA_SHA224,
+      CKM_ECDSA_SHA256,
+      CKM_ECDSA_SHA384,
+      CKM_ECDSA_SHA512,
+      CKM_ECDH1_DERIVE,
+      CKM_SHA_1,
+      CKM_SHA256,
+      CKM_SHA384,
+      CKM_SHA512};
 
   static const CK_MECHANISM_INFO token_mechanism_infos[] = { // KEEP ALIGNED WITH token_mechanisms
     {1024, 2048, CKF_HW | CKF_GENERATE_KEY_PAIR},
@@ -217,13 +217,14 @@ static void test_mechanism_list_and_info() {
     {1024, 2048, CKF_HW | CKF_SIGN | CKF_VERIFY},
     {1024, 2048, CKF_HW | CKF_SIGN | CKF_VERIFY},
     {1024, 2048, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_GENERATE_KEY_PAIR},
-    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY},
-    {256, 384, CKF_HW | CKF_DERIVE},
+    {256, 384, CKF_HW | CKF_GENERATE_KEY_PAIR | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_SIGN | CKF_VERIFY | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
+    {256, 384, CKF_HW | CKF_DERIVE | CKF_EC_F_P | CKF_EC_NAMEDCURVE | CKF_EC_UNCOMPRESS},
     {0, 0, CKF_DIGEST},
     {0, 0, CKF_DIGEST},
     {0, 0, CKF_DIGEST},
@@ -467,9 +468,8 @@ static void test_sign_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_ULON
   test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA, eck, key_len);
   test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA1, eck, key_len);
   test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA256, eck, key_len);
-  if(key_len > 32) {
-    test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA384, eck, key_len);
-  }
+  test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA384, eck, key_len);
+  test_ec_sign_thorough(funcs, session, obj_pvtkey, CKM_ECDSA_SHA512, eck, key_len);
 
   EC_KEY_free(eck);
   destroy_test_objects(funcs, session, obj_pvtkey, N_SELECTED_KEYS);
@@ -508,6 +508,57 @@ static void test_generate_rsa(CK_ULONG key_size, CK_BYTE n_keys) {
   asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
 }
 
+static void test_import_rsa_with_policy() {
+  dprintf(0, "TEST START: test_import_rsa_with_policy()\n");
+  CK_ULONG key_size = 2048;
+  CK_SESSION_HANDLE session;
+  CK_BYTE i, j;
+  CK_BYTE touch_policies[] = {YKPIV_TOUCHPOLICY_ALWAYS, YKPIV_TOUCHPOLICY_CACHED, YKPIV_TOUCHPOLICY_NEVER, YKPIV_TOUCHPOLICY_DEFAULT};
+  CK_BYTE pin_policies[] = {YKPIV_PINPOLICY_ALWAYS, YKPIV_PINPOLICY_ONCE, YKPIV_PINPOLICY_NEVER, YKPIV_PINPOLICY_DEFAULT};
+
+  init_connection();
+  asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
+  asrt(funcs->C_Login(session, CKU_SO, (CK_CHAR_PTR)"010203040506070801020304050607080102030405060708", 48), CKR_OK, "Login SO");
+  for (i = 0; i < sizeof(touch_policies)/sizeof(touch_policies[0]); i++) {
+    for (j = 0; j < sizeof(pin_policies)/sizeof(pin_policies[0]); j++) { 
+      import_rsa_key_with_policy(funcs, session, key_size, is_neo ? N_SELECTED_KEYS : N_ALL_KEYS, 
+                                 touch_policies[i], pin_policies[j], CK_FALSE);
+    }
+  }
+
+  import_rsa_key_with_policy(funcs, session, key_size, is_neo ? N_SELECTED_KEYS : N_ALL_KEYS,
+                             YKPIV_TOUCHPOLICY_DEFAULT, YKPIV_PINPOLICY_DEFAULT, CK_TRUE);
+
+  asrt(funcs->C_Logout(session), CKR_OK, "Logout SO");
+  asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
+  asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
+  dprintf(0, "TEST END: test_import_rsa_with_policy()\n");
+}
+
+static void test_generate_eccp256_with_policy() {
+  dprintf(0, "TEST START: test_generate_eccp256_with_policy()\n");
+  CK_SESSION_HANDLE session;
+  CK_BYTE i, j;
+  CK_BYTE touch_policies[] = {YKPIV_TOUCHPOLICY_ALWAYS, YKPIV_TOUCHPOLICY_CACHED, YKPIV_TOUCHPOLICY_NEVER, YKPIV_TOUCHPOLICY_DEFAULT};
+  CK_BYTE pin_policies[] = {YKPIV_PINPOLICY_ALWAYS, YKPIV_PINPOLICY_ONCE, YKPIV_PINPOLICY_NEVER, YKPIV_PINPOLICY_DEFAULT};  
+  CK_BYTE  params_eccp256[] = {0x06, 0x08, 0x2a, 0x86, 0x48, 0xce, 0x3d, 0x03, 0x01, 0x07};
+
+  init_connection();
+  asrt(funcs->C_OpenSession(0, CKF_SERIAL_SESSION | CKF_RW_SESSION, NULL, NULL, &session), CKR_OK, "OpenSession1");
+  for (i = 0; i < sizeof(touch_policies)/sizeof(touch_policies[0]); i++) {
+    for (j = 0; j < sizeof(pin_policies)/sizeof(pin_policies[0]); j++) {
+      generate_ec_keys_with_policy(funcs, session, is_neo ? N_SELECTED_KEYS : N_ALL_KEYS, params_eccp256,
+                                   sizeof(params_eccp256), touch_policies[i], pin_policies[j], CK_FALSE);
+    }
+  }
+
+  generate_ec_keys_with_policy(funcs, session, is_neo ? N_SELECTED_KEYS : N_ALL_KEYS, params_eccp256, sizeof(params_eccp256),
+                               YKPIV_TOUCHPOLICY_DEFAULT, YKPIV_PINPOLICY_DEFAULT, CK_TRUE);
+  asrt(funcs->C_CloseSession(session), CKR_OK, "CloseSession");
+  asrt(funcs->C_Finalize(NULL), CKR_OK, "FINALIZE");
+  dprintf(0, "TEST END: test_generate_eccp256_with_policy()\n");
+}
+
 static void test_generate_rsakeys() {
   dprintf(0, "TEST START: test_generate_rsakeys()\n");
   if(is_neo) {
@@ -634,28 +685,6 @@ static void test_find_objects() {
   dprintf(0, "TEST END: test_find_objects()\n");
 }
 
-static CK_OBJECT_HANDLE get_public_key_handle(CK_SESSION_HANDLE session, CK_OBJECT_HANDLE privkey) {
-  CK_OBJECT_HANDLE found_obj[10] = {0};
-  CK_ULONG n_found_obj = 0;
-  CK_ULONG class_pub = CKO_PUBLIC_KEY;
-  CK_BYTE ckaid = 0;
-
-  CK_ATTRIBUTE idTemplate[] = {
-    {CKA_ID, &ckaid, sizeof(ckaid)}
-  };
-  CK_ATTRIBUTE idClassTemplate[] = {
-    {CKA_ID, &ckaid, sizeof(ckaid)},
-    {CKA_CLASS, &class_pub, sizeof(class_pub)}
-  };
-
-  asrt(funcs->C_GetAttributeValue(session, privkey, idTemplate, 1), CKR_OK, "GET CKA_ID");
-  asrt(funcs->C_FindObjectsInit(session, idClassTemplate, 2), CKR_OK, "FIND INIT");
-  asrt(funcs->C_FindObjects(session, found_obj, 10, &n_found_obj), CKR_OK, "FIND");
-  asrt(n_found_obj, 1, "N FOUND OBJS");
-  asrt(funcs->C_FindObjectsFinal(session), CKR_OK, "FIND FINAL");
-  return found_obj[0];
-}
-
 static void test_import_eccp(CK_BYTE* key_params, CK_ULONG key_params_len, CK_ULONG key_len, int curve, CK_BYTE n_keys) {
   EC_KEY            *eck;
   CK_OBJECT_HANDLE  obj_cert[N_ALL_KEYS]={0}, obj_pvtkey[N_ALL_KEYS]={0};
@@ -842,13 +871,15 @@ static void test_digest() {
 
 #endif
 
-int destruction_confirmed(void) {
+#if HW_TESTS
+static int destruction_confirmed(void) {
 #ifdef _WIN32
   return 1;
 #else
   return system("../../../tools/confirm.sh") == 0;
 #endif
 }
+#endif
 
 int main(void) {
 
@@ -879,9 +910,11 @@ int main(void) {
   test_generate_eccp256();
   test_generate_eccp384();
   test_generate_rsakeys();
+  test_generate_eccp256_with_policy();
   test_import_eccp256();
   test_import_eccp384();
   test_import_rsakeys();
+  test_import_rsa_with_policy();
   test_sign_eccp256();
   test_sign_eccp384();
   test_sign_rsakeys();
